@@ -110,8 +110,11 @@ case $VENV in
     # Make the workspace owned by the stack user
     sudo chown -R $STACK_USER:$STACK_USER $BASE
 
+    # temporary fix for bug 1693689
+    export IPV4_ADDRS_SAFE_TO_USE=${DEVSTACK_GATE_IPV4_ADDRS_SAFE_TO_USE:-${DEVSTACK_GATE_FIXED_RANGE:-10.1.0.0/20}}
+
     # deploy devstack as per local.conf
-    cd $DEVSTACK_PATH && sudo -H -u $GATE_STACK_USER ./stack.sh
+    cd $DEVSTACK_PATH && sudo -E -H -u $GATE_STACK_USER ./stack.sh
     ;;
 
 "tempest")
