@@ -89,11 +89,11 @@ latex_documents = [
 # -- Options for oslo_config.sphinxconfiggen ---------------------------------
 
 _config_generator_config_files = [
+    'neutron-agent.conf',
     'bagpipe-bgp.conf',
     'dataplane-evpn-linux-vxlan.conf',
     'dataplane-ipvpn-mpls-linux.conf',
     'dataplane-ipvpn-mpls-ovs.conf',
-    'ml2-bagpipe.conf',
 ]
 
 def _get_config_generator_config_definition(config_file):
