@@ -23,6 +23,8 @@ from oslo_config import cfg
 from oslo_log import log as logging
 import pbr.version
 
+from neutron.common import config as n_config  # noqa
+
 from networking_bagpipe.bagpipe_bgp.api import api
 from networking_bagpipe.bagpipe_bgp.common import config
 from networking_bagpipe.bagpipe_bgp.engine import exabgp_peer_worker
@@ -51,7 +53,6 @@ def setup_logging():
         '%s.engine.exabgp_peer_worker.exabgp=INFO' % BAGPIPE_BGP_MODULE,
         '%s.common.looking_glass=WARNING' % BAGPIPE_BGP_MODULE,
         '%s.engine.route_table_manager=INFO' % BAGPIPE_BGP_MODULE,
-        '%s.engine.tracker_worker=INFO' % BAGPIPE_BGP_MODULE
     ]
 
     logging.set_defaults(default_log_levels=(logging.get_default_log_levels() +
