@@ -25,6 +25,7 @@ import six
 import stevedore
 
 from networking_bagpipe.bagpipe_bgp.common import config
+from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
 from networking_bagpipe.bagpipe_bgp.common import log_decorator
 from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
 from networking_bagpipe.bagpipe_bgp.common import run_command
@@ -123,7 +124,6 @@ class DataplaneDriver(lg.LookingGlassLocalLogger):
     dataplane_instance_class = object  # has to be overridden by subclasses
     encaps = [exa.Encapsulation(exa.Encapsulation.Type.DEFAULT),
               exa.Encapsulation(exa.Encapsulation.Type.MPLS)]
-    makebefore4break_support = False
     ecmp_support = False
     required_kernel = None
 
@@ -225,6 +225,28 @@ class DataplaneDriver(lg.LookingGlassLocalLogger):
     def supported_encaps(self):
         return self.__class__.encaps
 
+    def needs_cleanup_assist(self):
+        '''Control per-route cleanup events
+
+        A dataplane driver not able to cleanup all the states for a given
+        VPN instance can return True here to receive artifical dataplane
+        removal calls, such as remove_dataplane_for_remote_endpoint, for
+        each state previously setup
+        '''
+        return False
+
+    def validate_directions(self, direction):
+        # by default, assume a driver only supports plugging in both directions
+        # if a driver does not support forwarding only the traffic to the port,
+        # (and hence omit forwarding traffic from the port), it should raise an
+        # exception if directions is TO_PORT
+        # if a driver does not support forwarding only the traffic from the
+        # port (and hence omit forwarding traffic to the port), it should raise
+        # an exception if directions is FROM_PORT
+        if (direction is not None) and (direction != constants.BOTH):
+            self.log.warning("Unsupported direction: %s", direction)
+            raise exc.APIException("Unsupported direction: %s" % direction)
+
     def _run_command(self, command, run_as_root=False, *args, **kwargs):
         return run_command.run_command(self.log, command, run_as_root,
                                        *args, **kwargs)
@@ -262,12 +284,13 @@ class VPNInstanceDataplane(lg.LookingGlassLocalLogger):
         pass
 
     @abc.abstractmethod
-    def vif_plugged(self, mac_address, ip_address_prefix, localport, label):
+    def vif_plugged(self, mac_address, ip_address_prefix, localport, label,
+                    direction):
         pass
 
     @abc.abstractmethod
     def vif_unplugged(self, mac_address, ip_address_prefix, localport, label,
-                      last_endpoint=True):
+                      direction, last_endpoint=True):
         pass
 
     def update_fallback(self, fallback):
@@ -287,6 +310,16 @@ class VPNInstanceDataplane(lg.LookingGlassLocalLogger):
                                              lb_consistent_hash_order=0):
         pass
 
+    def needs_cleanup_assist(self):
+        '''Control per-route cleanup events
+
+        A dataplane driver not able to cleanup all the states for a given
+        VPN instance can return True here to receive artifical dataplane
+        removal calls, such as remove_dataplane_for_remote_endpoint, for
+        each state previously setup
+        '''
+        return self.driver.needs_cleanup_assist()
+
     def _run_command(self, command, run_as_root=False, *args, **kwargs):
         return self.driver._run_command(command, run_as_root, *args, **kwargs)
 
@@ -308,28 +341,27 @@ class DummyVPNInstanceDataplane(VPNInstanceDataplane):
         VPNInstanceDataplane.__init__(self, *args)
 
     @log_decorator.log
-    def vif_plugged(self, mac_address, ip_address_prefix, localport, label):
+    def vif_plugged(self, *args, **kwargs):
         pass
 
     @log_decorator.log
-    def vif_unplugged(self, mac_address, ip_address_prefix, localport, label,
-                      last_endpoint=True):
+    def vif_unplugged(self, *args, **kwargs):
         pass
 
     @log_decorator.log
-    def update_fallback(self, fallback):
+    def update_fallback(self, *args, **kwargs):
         pass
 
     @log_decorator.log
-    def setup_dataplane_for_remote_endpoint(self, *args):
+    def setup_dataplane_for_remote_endpoint(self, *args, **kwargs):
         pass
 
     @log_decorator.log
-    def remove_dataplane_for_remote_endpoint(self, *args):
+    def remove_dataplane_for_remote_endpoint(self, *args, **kwargs):
         pass
 
     @log_decorator.log
-    def cleanup(self):
+    def cleanup(self, *args, **kwargs):
         pass
 
 
@@ -342,9 +374,9 @@ class DummyDataplaneDriver(DataplaneDriver):
         self.log.warning("Dummy dataplane driver, won't do anything useful")
 
     @log_decorator.log_info
-    def initialize(self):
+    def initialize(self, *args, **kwargs):
         pass
 
     @log_decorator.log_info
-    def reset_state(self):
+    def reset_state(self, *args, **kwargs):
         pass
