@@ -33,11 +33,11 @@ from networking_bagpipe.bagpipe_bgp.vpn import vpn_instance
 class VPNInstanceDataplane(dp_drivers.VPNInstanceDataplane):
 
     @abc.abstractmethod
-    def add_dataplane_for_bum_endpoint(self, remote_pe, label, nlri, encaps):
+    def add_dataplane_for_bum_endpoint(self, remote_pe, dpid, nlri, encaps):
         pass
 
     @abc.abstractmethod
-    def remove_dataplane_for_bum_endpoint(self, remote_pe, label, nlri):
+    def remove_dataplane_for_bum_endpoint(self, remote_pe, dpid, nlri):
         pass
 
     @abc.abstractmethod
@@ -61,23 +61,23 @@ class VPNInstanceDataplane(dp_drivers.VPNInstanceDataplane):
 
 
 class DummyVPNInstanceDataplane(dp_drivers.DummyVPNInstanceDataplane,
-                                dp_drivers.VPNInstanceDataplane):
+                                VPNInstanceDataplane):
     '''Dummy, do-nothing dataplane driver'''
 
     @log_decorator.log_info
-    def add_dataplane_for_bum_endpoint(self, remote_pe, label, nlri, encaps):
+    def add_dataplane_for_bum_endpoint(self, *args, **kwargs):
         pass
 
     @log_decorator.log_info
-    def remove_dataplane_for_bum_endpoint(self, remote_pe, label, nlri):
+    def remove_dataplane_for_bum_endpoint(self, *args, **kwargs):
         pass
 
     @log_decorator.log_info
-    def set_gateway_port(self, linuxif, gateway_ip):
+    def set_gateway_port(self, *args, **kwargs):
         pass
 
     @log_decorator.log_info
-    def gateway_port_down(self, linuxif):
+    def gateway_port_down(self, *args, **kwargs):
         pass
 
 
@@ -88,8 +88,8 @@ class DummyDataplaneDriver(dp_drivers.DummyDataplaneDriver):
     dataplane_instance_class = DummyVPNInstanceDataplane
     encaps = [exa.Encapsulation(exa.Encapsulation.Type.VXLAN)]
 
-    def __init__(self, *args):
-        dp_drivers.DummyDataplaneDriver.__init__(self, *args)
+    def __init__(self, *args, **kwargs):
+        dp_drivers.DummyDataplaneDriver.__init__(self, *args, **kwargs)
 
 
 class EVI(vpn_instance.VPNInstance, lg.LookingGlassMixin):
@@ -109,6 +109,16 @@ class EVI(vpn_instance.VPNInstance, lg.LookingGlassMixin):
 
         self.gw_port = None
 
+        encaps = self.dp_driver.supported_encaps()
+        if (exa.Encapsulation(exa.Encapsulation.Type.VXLAN) in encaps
+                and any([
+                exa.Encapsulation(exa.Encapsulation.Type.MPLS) in encaps,
+                exa.Encapsulation(exa.Encapsulation.Type.GRE) in encaps,
+                exa.Encapsulation(exa.Encapsulation.Type.MPLS_UDP) in encaps
+                ])):
+            raise Exception("The dataplane can't support both an MPLS encap "
+                            "and a VXLAN encapsulation")
+
         # Advertise route to receive multi-destination traffic
         self.log.info("Generating BGP route for broadcast/multicast traffic")
 
@@ -126,22 +136,31 @@ class EVI(vpn_instance.VPNInstance, lg.LookingGlassMixin):
         # add PMSI Tunnel Attribute route
         attributes.add(
             exa.PMSIIngressReplication(self.dp_driver.get_local_address(),
-                                       self.instance_label))
+                                       raw_label=self.instance_label))
 
         self.multicast_route_entry = engine.RouteEntry(nlri, self.export_rts,
                                                        attributes)
 
         self._advertise_route(self.multicast_route_entry)
 
+    def _vxlan_dp_driver(self):
+        return (exa.Encapsulation(exa.Encapsulation.Type.VXLAN) in
+                self.dp_driver.supported_encaps())
+
     def generate_vif_bgp_route(self, mac_address, ip_prefix, plen, label, rd):
         # Generate BGP route and advertise it...
 
         assert(plen == 32)
 
+        if self._vxlan_dp_driver():
+            mpls_label_field = exa.Labels([], raw_labels=[self.instance_label])
+        else:
+            mpls_label_field = exa.Labels([self.instance_label])
+
         # label parameter ignored, we need to use instance label
         nlri = exa.EVPNMAC(
             rd, exa.ESI(), exa.EthernetTag(), exa.MAC(mac_address), 6*8,
-            exa.Labels([self.instance_label]),
+            mpls_label_field,
             exa.IP.create(ip_prefix), None,
             exa.IP.create(self.dp_driver.get_local_address()))
 
@@ -160,9 +179,25 @@ class EVI(vpn_instance.VPNInstance, lg.LookingGlassMixin):
     def has_gateway_port(self):
         return (self.gw_port is not None)
 
+    def _interpret_nlri_mpls_field(self, label_field):
+        # interpret the MPLS field of an EVPN MAroute, based on
+        # whether we assume this is a VXLAN VNI or not
+        if self._vxlan_dp_driver():
+            return label_field.raw_labels[0]
+        else:
+            return label_field.labels[0]
+
+    def _interpret_pta_label_field(self, pmsi_tunnel):
+        # interpret the MPLS field of a PMSI tunnel attribute, based on
+        # whether we assume this is a VXLAN VNI or not
+        if self._vxlan_dp_driver():
+            return pmsi_tunnel.raw_label
+        else:
+            return pmsi_tunnel.label
+
     # TrackerWorker callbacks for BGP route updates ##########################
 
-    def _route_2_tracked_entry(self, route):
+    def route_to_tracked_entry(self, route):
         if isinstance(route.nlri, exa.EVPNMAC):
             return (exa.EVPNMAC, route.nlri.mac)
         elif isinstance(route.nlri, exa.EVPNMulticast):
@@ -177,7 +212,7 @@ class EVI(vpn_instance.VPNInstance, lg.LookingGlassMixin):
 
     @utils.synchronized
     @log_decorator.log
-    def _new_best_route(self, entry, new_route):
+    def new_best_route(self, entry, new_route):
         (entry_class, info) = entry
 
         encaps = self._check_encaps(new_route)
@@ -189,10 +224,11 @@ class EVI(vpn_instance.VPNInstance, lg.LookingGlassMixin):
 
             remote_pe = new_route.nexthop
 
-            label = new_route.nlri.label.labels[0]
+            dataplane_id = self._interpret_nlri_mpls_field(
+                new_route.nlri.label)
 
             self.dataplane.setup_dataplane_for_remote_endpoint(
-                prefix, remote_pe, label, new_route.nlri, encaps)
+                prefix, remote_pe, dataplane_id, new_route.nlri, encaps)
 
         elif entry_class == exa.EVPNMulticast:
             remote_endpoint = info
@@ -205,19 +241,19 @@ class EVI(vpn_instance.VPNInstance, lg.LookingGlassMixin):
                                  type(pmsi_tunnel))
             else:
                 remote_endpoint = pmsi_tunnel.ip
-                label = pmsi_tunnel.label
+                dataplane_id = self._interpret_pta_label_field(pmsi_tunnel)
 
                 self.log.info("Setting up dataplane for new ingress "
                               "replication destination %s", remote_endpoint)
                 self.dataplane.add_dataplane_for_bum_endpoint(
-                    remote_endpoint, label, new_route.nlri, encaps)
+                    remote_endpoint, dataplane_id, new_route.nlri, encaps)
         else:
             self.log.warning("unsupported entry_class: %s",
                              entry_class.__name__)
 
     @utils.synchronized
     @log_decorator.log
-    def _best_route_removed(self, entry, old_route, last):
+    def best_route_removed(self, entry, old_route, last):
         (entry_class, info) = entry
 
         if entry_class == exa.EVPNMAC:
@@ -227,30 +263,52 @@ class EVI(vpn_instance.VPNInstance, lg.LookingGlassMixin):
                                "dataplane does not want it")
                 return
 
+            def ip_dpid_from_route(route):
+                return (route.nexthop,
+                        self._interpret_nlri_mpls_field(route.nlri.label))
+
+            if self.equivalent_route_in_best_routes(old_route,
+                                                    ip_dpid_from_route):
+                self.log.debug("Route for same dataplane is still in best "
+                               "routes, skipping removal")
+                return
+
             prefix = info
 
-            remote_pe = old_route.nexthop
-            label = old_route.nlri.label.labels[0]
+            remote_pe, dataplane_id = ip_dpid_from_route(old_route)
 
             self.dataplane.remove_dataplane_for_remote_endpoint(
-                prefix, remote_pe, label, old_route.nlri)
+                prefix, remote_pe, dataplane_id, old_route.nlri)
 
         elif entry_class == exa.EVPNMulticast:
             remote_endpoint = info
 
+            def ip_dpid_from_route(route):
+                pmsi_tunnel = route.attributes.get(exa.PMSI.ID)
+                remote_endpoint = pmsi_tunnel.ip
+                dataplane_id = self._interpret_pta_label_field(pmsi_tunnel)
+                return (remote_endpoint, dataplane_id)
+
             # check that the route is actually carrying an PMSITunnel of type
             # ingress replication
             pmsi_tunnel = old_route.attributes.get(exa.PMSI.ID)
             if not isinstance(pmsi_tunnel, exa.PMSIIngressReplication):
                 self.log.warning("PMSITunnel of suppressed route is of"
                                  " unsupported type")
-            else:
-                remote_endpoint = pmsi_tunnel.ip
-                label = pmsi_tunnel.label
-                self.log.info("Cleaning up dataplane for ingress replication "
-                              "destination %s", remote_endpoint)
-                self.dataplane.remove_dataplane_for_bum_endpoint(
-                    remote_endpoint, label, old_route.nlri)
+                return
+
+            if self.equivalent_route_in_best_routes(old_route,
+                                                    ip_dpid_from_route):
+                self.log.debug("Route for same dataplane is still in best "
+                               "routes, skipping removal")
+                return
+
+            remote_endpoint, dataplane_id = ip_dpid_from_route(old_route)
+            self.log.info("Cleaning up dataplane for ingress replication "
+                          "destination %s", remote_endpoint)
+
+            self.dataplane.remove_dataplane_for_bum_endpoint(
+                remote_endpoint, dataplane_id, old_route.nlri)
         else:
             self.log.warning("unsupported entry_class: %s",
                              entry_class.__name__)
