@@ -16,6 +16,7 @@
 # limitations under the License.
 
 from oslo_config import cfg
+import pyroute2
 
 from networking_bagpipe.bagpipe_bgp.common import log_decorator
 from networking_bagpipe.bagpipe_bgp import constants as consts
@@ -40,9 +41,6 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
                 BRIDGE_NAME_PREFIX +
                 self.external_instance_id)[:consts.LINUX_DEV_LEN]
 
-        self.vxlan_if_name = (VXLAN_INTERFACE_PREFIX +
-                              self.external_instance_id)[:consts.LINUX_DEV_LEN]
-
         self.log.info("EVI %d: Initializing bridge %s",
                       self.instance_id, self.bridge_name)
         if not self._interface_exists(self.bridge_name):
@@ -74,6 +72,10 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
 
         self._cleaning_up = True
 
+        # removing the vxlan interface removes our routes,
+        # but if we don't remove the vxlan if (if it was reused) then
+        # cleanup will not happen, which is why we use cleanup assist
+        # (see needs cleanup assist below)
         self._cleanup_vxlan_if()
 
         # Delete only EVPN Bridge (Created by dataplane driver)
@@ -85,7 +87,28 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
                               run_as_root=True,
                               raise_on_error=False)
 
+    def needs_cleanup_assist(self):
+        # If we reused a vxlan interface we won't cleanup fdb entries
+        # in cleanup(), so we need to have remove_dataplane_for_x
+        # be called for reach state via cleanup assist
+        return VXLAN_INTERFACE_PREFIX not in self.vxlan_if_name
+
     def _create_and_plug_vxlan_if(self):
+        # if a VXLAN interface, with the VNI we want to use, is already plugged
+        # in the bridge, we want to reuse it
+        with pyroute2.IPDB(plugins=('interfaces',)) as ipdb:
+            for port_id in ipdb.interfaces[self.bridge_name].ports:
+                port = ipdb.interfaces[port_id]
+                if (port.kind == "vxlan" and
+                        port.vxlan_id == self.instance_label):
+                    self.log.info("reuse vxlan interface %s for VXLAN VNI %s",
+                                  port.ifname, self.instance_label)
+                    self.vxlan_if_name = port.ifname
+                    return
+
+        self.vxlan_if_name = (VXLAN_INTERFACE_PREFIX +
+                              self.external_instance_id)[:consts.LINUX_DEV_LEN]
+
         self.log.debug("Creating and plugging VXLAN interface %s",
                        self.vxlan_if_name)
 
@@ -114,6 +137,10 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
                           run_as_root=True)
 
     def _cleanup_vxlan_if(self):
+        if VXLAN_INTERFACE_PREFIX not in self.vxlan_if_name:
+            self.log.debug("we reused the VXLAN interface, don't cleanup")
+            return
+
         if self._is_vxlan_if_on_bridge():
             # Unplug VXLAN interface from Linux bridge
             self._unplug_from_bridge(self.vxlan_if_name)
@@ -121,20 +148,26 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
         self._remove_vxlan_if()
 
     def _remove_vxlan_if(self):
+        if not VXLAN_INTERFACE_PREFIX not in self.vxlan_if_name:
+            self.log.debug("we reused the VXLAN interface, don't remove")
+            return
+
         # Remove VXLAN interface
         self._run_command("ip link set %s down" % self.vxlan_if_name,
                           run_as_root=True)
         self._run_command("ip link del %s" % self.vxlan_if_name,
                           run_as_root=True)
 
-    def _is_vxlan_if_on_bridge(self):
-        (output, _) = self._run_command(
-            "brctl show %s | grep '%s' | sed -e 's/\s\+//g'" %
-            (self.bridge_name, VXLAN_INTERFACE_PREFIX),
-            run_as_root=True,
-            shell=True)
+    def _is_if_on_bridge(self, ifname):
+        with pyroute2.IPDB(plugins=('interfaces',)) as ipdb:
+            for port_id in ipdb.interfaces[self.bridge_name].ports:
+                port = ipdb.interfaces[port_id]
+                if port.ifname == ifname:
+                    return True
+        return False
 
-        return True if (output == self.vxlan_if_name) else False
+    def _is_vxlan_if_on_bridge(self):
+        return self._is_if_on_bridge(self.vxlan_if_name)
 
     def _interface_exists(self, interface):
         """Check if interface exists."""
@@ -151,22 +184,12 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
                               acceptable_return_codes=[0, 1])
 
     def set_gateway_port(self, linuxif, gw_ip):
-        gw_mac = "01:00:00:00:00:00"  # FIXME
-
         self._run_command("brctl addif %s %s" %
                           (self.bridge_name, linuxif),
                           run_as_root=True,
                           raise_on_error=False)
 
-        self._run_command("bridge fdb replace %s dev %s" %
-                          (gw_mac, linuxif),
-                          run_as_root=True)
-
-        self._run_command(
-            "ip neighbor replace %s lladdr %s dev %s nud permanent" %
-            (gw_ip, gw_mac, linuxif),
-            run_as_root=True
-        )
+        self._fdb_dump()
 
     def gateway_port_down(self, linuxif):
         self._run_command("brctl delif %s %s" %
@@ -179,7 +202,7 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
         self.bridge_name = linuxbr
 
     @log_decorator.log_info
-    def vif_plugged(self, mac_address, ip_address, localport, label):
+    def vif_plugged(self, mac_address, ip_address, localport, dpid, direction):
         # Plug localport only if bridge was created by us
         if BRIDGE_NAME_PREFIX in self.bridge_name:
             self.log.debug("Plugging localport %s into EVPN bridge %s",
@@ -193,17 +216,28 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
                           (mac_address, localport['linuxif']),
                           run_as_root=True)
 
+        self._fdb_dump()
+
     @log_decorator.log_info
-    def vif_unplugged(self, mac_address, ip_address, localport, label,
-                      last_endpoint=True):
+    def vif_unplugged(self, mac_address, ip_address, localport, dpid,
+                      direction, last_endpoint=True):
+
+        # remove local fdb entry, but only if tap interface is still here
+        if self._is_if_on_bridge(localport['linuxif']):
+            self._run_command("bridge fdb delete %s dev %s" %
+                              (mac_address, localport['linuxif']),
+                              run_as_root=True)
+
         # unplug localport only if bridge was created by us
         if BRIDGE_NAME_PREFIX in self.bridge_name:
             self.log.debug("Unplugging localport %s from EVPN bridge %s",
                            localport['linuxif'], self.bridge_name)
             self._unplug_from_bridge(localport['linuxif'])
 
+        self._fdb_dump()
+
     @log_decorator.log
-    def setup_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
+    def setup_dataplane_for_remote_endpoint(self, prefix, remote_pe, dpid,
                                             nlri, encaps):
         if self._cleaning_up:
             self.log.debug("setup_dataplane_for_remote_endpoint: instance"
@@ -212,7 +246,7 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
 
         mac = prefix
         ip = nlri.ip
-        vni = label
+        vni = dpid
 
         # populate bridge forwarding db
         self._run_command("bridge fdb replace %s dev %s dst %s vni %s" %
@@ -225,13 +259,12 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
                               "permanent" % (ip, mac, self.vxlan_if_name),
                               run_as_root=True)
         else:
-            self.log.warning("No IP in E-VPN route, ARP will not work for this"
-                             "IP/MAC")
+            self.log.trace("No IP in E-VPN route, no ARP proxy for %s" % mac)
 
         self._fdb_dump()
 
     @log_decorator.log
-    def remove_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
+    def remove_dataplane_for_remote_endpoint(self, prefix, remote_pe, dpid,
                                              nlri):
         if self._cleaning_up:
             self.log.debug("setup_dataplane_for_remote_endpoint: instance"
@@ -240,13 +273,16 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
 
         mac = prefix
         ip = nlri.ip
-        vni = label
+        vni = dpid
 
         self._fdb_dump()
 
-        self._run_command("ip neighbor del %s lladdr %s dev %s nud permanent" %
-                          (ip, mac, self.vxlan_if_name),
-                          run_as_root=True)
+        # clear ARP proxy
+        if ip is not None:
+            self._run_command("ip neighbor del %s lladdr %s dev %s nud "
+                              "permanent" % (ip, mac, self.vxlan_if_name),
+                              run_as_root=True)
+
         self._run_command("bridge fdb del %s dev %s dst %s vni %s" %
                           (mac, self.vxlan_if_name, remote_pe, vni),
                           run_as_root=True)
@@ -254,13 +290,13 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
         self._fdb_dump()
 
     @log_decorator.log
-    def add_dataplane_for_bum_endpoint(self, remote_pe, label, nlri, encaps):
+    def add_dataplane_for_bum_endpoint(self, remote_pe, dpid, nlri, encaps):
         if self._cleaning_up:
             self.log.debug("setup_dataplane_for_remote_endpoint: instance"
                            " cleaning up, do nothing")
             return
 
-        vni = label
+        vni = dpid
 
         # 00:00:00:00:00 usable as default since kernel commit
         # 58e4c767046a35f11a55af6ce946054ddf4a8580 (2013-06-25)
@@ -271,13 +307,13 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
         self._fdb_dump()
 
     @log_decorator.log
-    def remove_dataplane_for_bum_endpoint(self, remote_pe, label, nlri):
+    def remove_dataplane_for_bum_endpoint(self, remote_pe, dpid, nlri):
         if self._cleaning_up:
             self.log.debug("setup_dataplane_for_remote_endpoint: instance"
                            " cleaning up, do nothing")
             return
 
-        vni = label
+        vni = dpid
 
         self._fdb_dump()
 
@@ -290,7 +326,7 @@ class LinuxVXLANEVIDataplane(evpn.VPNInstanceDataplane):
     def _fdb_dump(self):
         if self.log.debug:
             self.log.debug("bridge fdb dump: %s", self._run_command(
-                "bridge fdb show dev %s" % self.vxlan_if_name,
+                "bridge fdb show br %s" % self.bridge_name,
                 run_as_root=True)[0])
 
     # Looking glass ####
@@ -312,7 +348,7 @@ class LinuxVXLANDataplaneDriver(dp_drivers.DataplaneDriver):
     encaps = [exa.Encapsulation(exa.Encapsulation.Type.VXLAN)]
 
     driver_opts = [
-        cfg.IntOpt("vxlan_dst_port", default="0",
+        cfg.IntOpt("vxlan_dst_port", default="4789",
                    help=("UDP port toward which send VXLAN traffic (defaults "
                          "to standard IANA-allocated port)")),
     ]
