@@ -15,6 +15,10 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import abc
+import itertools
+import six
+
 from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
 from networking_bagpipe.bagpipe_bgp.common import log_decorator
 from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
@@ -28,8 +32,32 @@ from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
 from networking_bagpipe.bagpipe_bgp.vpn import vpn_instance
 
 
+@six.add_metaclass(abc.ABCMeta)
+class VPNInstanceDataplane(dp_drivers.VPNInstanceDataplane):
+
+    @abc.abstractmethod
+    def add_dataplane_for_traffic_classifier(self, classifier,
+                                             redirect_to_instance_id):
+        pass
+
+    @abc.abstractmethod
+    def remove_dataplane_for_traffic_classifier(self, classifier):
+        pass
+
+
+class DummyVPNInstanceDataplane(dp_drivers.DummyVPNInstanceDataplane,
+                                VPNInstanceDataplane):
+
+    def add_dataplane_for_traffic_classifier(self, *args, **kwargs):
+        raise Exception("not implemented")
+
+    def remove_dataplane_for_traffic_classifier(self, *args, **kwargs):
+        raise Exception("not implemented")
+
+
 class DummyDataplaneDriver(dp_drivers.DummyDataplaneDriver):
     type = constants.IPVPN
+    dataplane_instance_class = DummyVPNInstanceDataplane
 
 
 class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
@@ -130,6 +158,19 @@ class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
         self.log.debug("RouteEntry for (re-)advertisement: %s", entry)
         return entry
 
+    @log_decorator.log
+    def _route_for_attract_static_dest_prefixes(self, label, rd):
+        if not self.attract_static_dest_prefixes:
+            return
+
+        for prefix in self.attract_static_dest_prefixes:
+            nlri = self._nlri_from(prefix, label, rd)
+
+            entry = engine.RouteEntry(nlri, self.readvertise_to_rts)
+            self.log.debug("RouteEntry for attract static destination prefix: "
+                           "%s", entry)
+            yield entry
+
     @log_decorator.log
     def _route_for_redirect_prefix(self, prefix):
         prefix_classifier = utils.dict_camelcase_to_underscore(
@@ -175,11 +216,11 @@ class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
         self.log.debug("Start re-advertising %s from VRF", nlri.cidr.prefix())
         for _, endpoints in self.localport_2_endpoints.items():
             for endpoint in endpoints:
-                port_data = self.mac_2_localport_data[endpoint['mac']]
+                port_data = self.mac_2_localport_data[endpoint[0]]
                 label = port_data['label']
                 lb_consistent_hash_order = port_data[
                     'lb_consistent_hash_order']
-                rd = self.endpoint_2_rd[(endpoint['mac'], endpoint['ip'])]
+                rd = self.endpoint_2_rd[endpoint]
                 self.log.debug("Start re-advertising %s from VRF, with label "
                                "%s and route distinguisher %s",
                                nlri, label, rd)
@@ -200,11 +241,11 @@ class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
         self.log.debug("Stop re-advertising %s from VRF", nlri.cidr.prefix())
         for _, endpoints in self.localport_2_endpoints.items():
             for endpoint in endpoints:
-                port_data = self.mac_2_localport_data[endpoint['mac']]
+                port_data = self.mac_2_localport_data[endpoint[0]]
                 label = port_data['label']
                 lb_consistent_hash_order = port_data[
                     'lb_consistent_hash_order']
-                rd = self.endpoint_2_rd[(endpoint['mac'], endpoint['ip'])]
+                rd = self.endpoint_2_rd[endpoint]
                 self.log.debug("Stop re-advertising %s from VRF, with label %s"
                                "and route distinguisher %s", nlri, label, rd)
                 self._withdraw_route_or_default(route, label, rd,
@@ -216,52 +257,58 @@ class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
 
         self.readvertised.remove(route)
 
+    @log_decorator.log_info
     def vif_plugged(self, mac_address, ip_address_prefix, localport,
-                    advertise_subnet=False, lb_consistent_hash_order=0):
-        vpn_instance.VPNInstance.vif_plugged(self, mac_address,
-                                             ip_address_prefix,
-                                             localport, advertise_subnet,
-                                             lb_consistent_hash_order)
+                    advertise_subnet=False, lb_consistent_hash_order=0,
+                    local_pref=None, **kwargs):
+        super(VRF, self).vif_plugged(mac_address, ip_address_prefix,
+                                     localport, advertise_subnet,
+                                     lb_consistent_hash_order, local_pref,
+                                     **kwargs)
 
-        label = self.mac_2_localport_data[mac_address]['label']
-        rd = self.endpoint_2_rd[(mac_address, ip_address_prefix)]
-        for route in self.readvertised:
-            self.log.debug("Re-advertising %s with this port as next hop",
-                           route.nlri)
-            self._advertise_route_or_default(route, label, rd,
-                                             lb_consistent_hash_order)
+        if vpn_instance.forward_to_port(kwargs.get('direction')):
+            endpoint = (mac_address, ip_address_prefix)
+            label = self.mac_2_localport_data[mac_address]['label']
+            rd = self.endpoint_2_rd[endpoint]
+            for route in itertools.chain(
+                    self.readvertised,
+                    self._route_for_attract_static_dest_prefixes(label, rd)):
+                self.log.debug("Re-advertising %s with this port as next hop",
+                               route.nlri)
+                self._advertise_route_or_default(route, label, rd,
+                                                 lb_consistent_hash_order)
 
-            if self.attract_traffic:
-                flow_route = self._route_for_redirect_prefix(
-                    route.nlri.cidr.prefix())
-                self._advertise_route(flow_route)
+                if self.attract_traffic:
+                    flow_route = self._route_for_redirect_prefix(
+                        route.nlri.cidr.prefix())
+                    self._advertise_route(flow_route)
 
-    def vif_unplugged(self, mac_address, ip_address_prefix,
-                      advertise_subnet=False,
-                      lb_consistent_hash_order=0):
-        label = self.mac_2_localport_data[mac_address]['label']
-        lb_consistent_hash_order = (self.mac_2_localport_data[mac_address]
-                                    ["lb_consistent_hash_order"])
-        rd = self.endpoint_2_rd[(mac_address, ip_address_prefix)]
-        for route in self.readvertised:
-            self.log.debug("Stop re-advertising %s with this port as next hop",
-                           route.nlri)
-            self._withdraw_route_or_default(route, label, rd,
-                                            lb_consistent_hash_order)
+    @log_decorator.log_info
+    def vif_unplugged(self, mac_address, ip_address_prefix):
+        endpoint = (mac_address, ip_address_prefix)
+        direction = self.endpoint_2_direction[endpoint]
+        if vpn_instance.forward_to_port(direction):
+            label = self.mac_2_localport_data[mac_address]['label']
+            lb_consistent_hash_order = (self.mac_2_localport_data[mac_address]
+                                        ["lb_consistent_hash_order"])
+            rd = self.endpoint_2_rd[endpoint]
+            for route in itertools.chain(
+                    self.readvertised,
+                    self._route_for_attract_static_dest_prefixes(label, rd)):
+                self.log.debug("Stop re-advertising %s", route.nlri)
+                self._withdraw_route_or_default(route, label, rd,
+                                                lb_consistent_hash_order)
 
-            if self.attract_traffic and self.has_only_one_endpoint():
-                flow_route = self._route_for_redirect_prefix(
-                    route.nlri.cidr.prefix())
-                self._withdraw_route(flow_route)
+                if self.attract_traffic and self.has_only_one_endpoint():
+                    flow_route = self._route_for_redirect_prefix(
+                        route.nlri.cidr.prefix())
+                    self._withdraw_route(flow_route)
 
-        vpn_instance.VPNInstance.vif_unplugged(self, mac_address,
-                                               ip_address_prefix,
-                                               advertise_subnet,
-                                               lb_consistent_hash_order)
+        super(VRF, self).vif_unplugged(mac_address, ip_address_prefix)
 
     # Callbacks for BGP route updates (TrackerWorker) ########################
 
-    def _route_2_tracked_entry(self, route):
+    def route_to_tracked_entry(self, route):
         if isinstance(route.nlri, ipvpn_routes.IPVPN):
             return route.nlri.cidr.prefix()
         elif isinstance(route.nlri, flowspec.Flow):
@@ -273,7 +320,7 @@ class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
 
     @utils.synchronized
     @log_decorator.log
-    def _new_best_route(self, entry, new_route):
+    def new_best_route(self, entry, new_route):
 
         if isinstance(new_route.nlri, flowspec.Flow):
             if len(new_route.ecoms(exa.TrafficRedirect)) == 1:
@@ -320,7 +367,7 @@ class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
 
     @utils.synchronized
     @log_decorator.log
-    def _best_route_removed(self, entry, old_route, last):
+    def best_route_removed(self, entry, old_route, last):
 
         if isinstance(old_route.nlri, flowspec.Flow):
             if len(old_route.ecoms(exa.TrafficRedirect)) == 1:
@@ -345,16 +392,37 @@ class VRF(vpn_instance.VPNInstance, lg.LookingGlassMixin):
                     self.log.debug("Need to stop re-advertising %s", prefix)
                     self._readvertise_stop(old_route)
 
-            if not self._imported(old_route):
-                self.log.debug("No need to update dataplane for:%s",
-                               prefix)
-                return
+            # NOTE(tmorin): On new best routes, we only trigger dataplane
+            # update events after checking with self._imported(...) that the
+            # route was imported (and not a route that we receive because the
+            # VRF should readvertise ir). On best_route_removed, we can't do
+            # that because we could end up in a situation where:
+            # - initially import_rts contains RT X
+            # - we receive a route for RT X and install dataplane state
+            # - the import_rts list is later updated and RT X is not anymore
+            # part of the imported RTs, and the VRF unsubscribes from RT X
+            # - we receive the best_route_removed callbacks corresponding to
+            # the unsubscribe, but since the route is for no RT that is in
+            # import_rts, we don't update the dataplane
+            # The result would be to fail to remove dataplane state for this
+            # route, so we're better not optimizing this case and remove
+            # dataplane state, including possibly for routes that we did
+            # not install in it.
 
             if self._skip_route_removal(last):
                 self.log.debug("Skipping removal of non-last route because "
                                "dataplane does not want it")
                 return
 
+            # if we still have a route with same dataplane properties in
+            # best routes, then we don't want to clear the dataplane entry
+            if self.equivalent_route_in_best_routes(
+                    old_route,
+                    lambda r: (r.nexthop, r.nlri.labels.labels[0])):
+                self.log.debug("Route for same dataplane is still in best "
+                               "routes, skipping removal")
+                return
+
             encaps = self._check_encaps(old_route)
             if not encaps:
                 return
