@@ -140,7 +140,8 @@ class MPLSLinuxVRFDataplane(dp_drivers.VPNInstanceDataplane):
         self.flush_routes()
 
     @log_decorator.log_info
-    def vif_plugged(self, mac_address, ip_address, localport, label):
+    def vif_plugged(self, mac_address, ip_address, localport, label,
+                    direction):
 
         interface = localport['linuxif']
 
@@ -180,10 +181,9 @@ class MPLSLinuxVRFDataplane(dp_drivers.VPNInstanceDataplane):
         except netlink.exceptions.NetlinkError as x:
             if x.code == errno.EEXIST:
                 # the route already exists, fine
-                self.log.warning("route %s already exists on %s",
-                                 self.gateway_ip,
-                                 interface)
-                pass
+                self.log.debug("route %s already exists on %s",
+                               self.gateway_ip,
+                               interface)
             else:
                 raise
 
@@ -202,13 +202,12 @@ class MPLSLinuxVRFDataplane(dp_drivers.VPNInstanceDataplane):
             if x.code == errno.EEXIST:
                 # the route already exists, fine
                 self.log.warning("MPLS state for %d already exists", label)
-                pass
             else:
                 raise
 
     @log_decorator.log_info
     def vif_unplugged(self, mac_address, ip_address, localport, label,
-                      last_endpoint=True):
+                      direction, last_endpoint=True):
         interface = localport['linuxif']
 
         if interface not in self.ip.interfaces:
