@@ -15,6 +15,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import operator
 import re
 
 from distutils import version  # pylint: disable=no-name-in-module
@@ -29,19 +30,24 @@ from networking_bagpipe.bagpipe_bgp.common import net_utils
 from networking_bagpipe.bagpipe_bgp import constants as consts
 from networking_bagpipe.bagpipe_bgp.engine import exa
 from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
+from networking_bagpipe.bagpipe_bgp.vpn import vpn_instance
 
+# man ovs-ofctl /32768
+DEFAULT_OVS_FLOW_PRIORITY = 0x8000
 
-DEFAULT_RULE_PRIORITY = 40000
+# we want to avoid having our flows having a lowest priority than
+# the default
+DEFAULT_RULE_PRIORITY = DEFAULT_OVS_FLOW_PRIORITY + 0x1000
 
-DEFAULT_ARPNS_IF_MTU = 9000
+# priorities for IP match flows
+# highest priority MAX_PREFIX_PRIORITY for MAX_PREFIX_LENGTH prefix
+# (MAX_PREFIX_PRIORITY - MAX_PREFIX_LENGTH) for a zero length prefix
+MAX_PREFIX_PRIORITY = DEFAULT_RULE_PRIORITY
+MAX_PREFIX_LENGTH = 0x80  # 128
 
-OVSBR2ARPNS_INTERFACE_PREFIX = "toarpns"
-
-# name of the veth device in the ARP proxy network namespace,
-#  whose remote end is plugged in the OVS bridge
-PROXYARP2OVS_IF = "ovs"
-
-ARPNETNS_PREFIX = "arp-vrf"
+# fallback flows get a priority even lower
+# (we round it for better readability of flow dumps)
+FALLBACK_PRIORITY = MAX_PREFIX_PRIORITY - MAX_PREFIX_LENGTH - 0x80
 
 NO_MPLS_PHY_INTERFACE = -1
 
@@ -59,27 +65,44 @@ OVS_DUMP_FLOW_FILTER = "| grep -v NXST_FLOW | perl -pe '"               \
 
 GATEWAY_MAC = "00:00:5e:00:43:64"
 
+ARP_RESPONDER_ACTIONS = ('move:NXM_OF_ETH_SRC[]->NXM_OF_ETH_DST[],'
+                         'mod_dl_src:%(mac)s,'
+                         'load:0x2->NXM_OF_ARP_OP[],'
+                         'move:NXM_NX_ARP_SHA[]->NXM_NX_ARP_THA[],'
+                         'push:NXM_OF_ARP_TPA[],'
+                         'move:NXM_OF_ARP_SPA[]->NXM_OF_ARP_TPA[],'
+                         'load:%(mac)#x->NXM_NX_ARP_SHA[],'
+                         'pop:NXM_OF_ARP_SPA[],'
+                         '%(vlan_action)soutput:%(in_port)d')
+
 VRF_REGISTER = 0
 LB_HOP_REGISTER = 1
 
 
-def get_ovsbr2arpns_if(namespace_id):
-    i = namespace_id.replace(ARPNETNS_PREFIX, "")
-    return (OVSBR2ARPNS_INTERFACE_PREFIX + i)[:consts.LINUX_DEV_LEN]
-
-
 def join_s(*args):
     return ','.join([_f for _f in args if _f])
 
 
+def _match_from_prefix(prefix):
+    # A zero-length prefix is a default route, no nw_dst is needed/possible
+    # in this case
+    prefix_length = netaddr.IPNetwork(prefix).prefixlen
+    return 'nw_dst=%s' % prefix if prefix_length != 0 else None
+
+
+def _priority_from_prefix(prefix):
+    prefix_length = netaddr.IPNetwork(prefix).prefixlen
+    # to implement a longest-match lookup we give longest prefixes
+    # the higher priority
+    priority = MAX_PREFIX_PRIORITY - (MAX_PREFIX_LENGTH - prefix_length)
+    return priority
+
+
 class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
 
     def __init__(self, *args, **kwargs):
         super(MPLSOVSVRFDataplane, self).__init__(*args, **kwargs)
 
-        self.arp_netns = ("%s%d" % (ARPNETNS_PREFIX,
-                                    self.instance_id))[:consts.LINUX_DEV_LEN]
-
         # Initialize dict where we store info on OVS ports (port numbers and
         # bound IP address)
         self._ovs_port_info = dict()
@@ -94,83 +117,6 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
         self.fallback = None
         self.push_vlan_action = None
 
-        if self.driver.proxy_arp:
-            self._init_arp_netns()
-
-    def _init_arp_netns(self):
-        self.log.info("VRF %d: Initializing network namespace %s for ARP "
-                      "proxing", self.instance_id, self.arp_netns)
-        # Get names of veth pair devices between OVS and network namespace
-
-        ovsbr_to_proxyarp_ns = get_ovsbr2arpns_if(self.arp_netns)
-
-        if not self._arp_net_ns_exists():
-            self.log.debug("VRF network namespace doesn't exist, creating...")
-            # Create network namespace
-            self._run_command("ip netns add %s" % self.arp_netns,
-                              run_as_root=True)
-
-            # Set up veth pair devices between OVS and ARP network namespace
-            self._create_arp_netns_veth_pair(ovsbr_to_proxyarp_ns,
-                                             PROXYARP2OVS_IF)
-
-            # Force MAC address of netns-to-OVS port
-            net_utils.set_device_mac(self._run_command,
-                                     PROXYARP2OVS_IF,
-                                     GATEWAY_MAC,
-                                     self.arp_netns)
-
-            # Retrieve broadcast IP address
-            ip = netaddr.IPNetwork("%s/%s" % (self.gateway_ip, self.mask))
-            broadcast_ip = str(ip.broadcast)
-
-            # Set up network namespace interface as gateway
-            self._run_command("ip netns exec %s ip addr add %s/%s broadcast %s"
-                              " dev %s" %
-                              (self.arp_netns, self.gateway_ip,
-                               self.mask, broadcast_ip, PROXYARP2OVS_IF),
-                              run_as_root=True,
-                              raise_on_error=True)
-
-            # Setup IP forwarding
-            self._run_command("ip netns exec %s sh -c \"echo 1 > /proc/sys"
-                              "/net/ipv4/ip_forward\"" % self.arp_netns,
-                              run_as_root=True,
-                              shell=True)
-            self._run_command("ip netns exec %s sh -c \"echo 1 > /proc/sys/net"
-                              "/ipv4/conf/all/forwarding\"" % self.arp_netns,
-                              run_as_root=True,
-                              shell=True)
-
-            # Setup ARP proxying
-            self._run_command("ip netns exec %s sh -c \"echo 1 > /proc/sys/net"
-                              "/ipv4/conf/%s/proxy_arp\"" %
-                              (self.arp_netns, PROXYARP2OVS_IF),
-                              run_as_root=True,
-                              shell=True)
-            self._run_command("ip netns exec %s sh -c \"echo 1 > /proc/sys/net"
-                              "/ipv4/conf/%s/proxy_arp_pvlan\"" %
-                              (self.arp_netns, PROXYARP2OVS_IF),
-                              run_as_root=True,
-                              shell=True)
-        else:
-            self.log.debug("VRF network namespace already exists...")
-
-        # OVS port number for the port toward the proxy ARP netns
-        self.arp_net_nsport = self.driver.find_ovs_port(ovsbr_to_proxyarp_ns)
-
-        # Map IP traffic from plugged ports to ARP/GW netns
-        self._ovs_flow_add(self._vrf_match('nw_dst=%s' % self.gateway_ip),
-                           'output:%s' % self.arp_net_nsport,
-                           self.driver.vrf_table)
-
-        # Map ARP traffic from plugged port to ARP/GW netns
-        # (rule to flow ARP response from GW to plugged port is added in
-        # vif_plugged)
-        self._ovs_flow_add(self._vrf_match(None, proto="arp"),
-                           'output:%s' % self.arp_net_nsport,
-                           self.driver.vrf_table)
-
     @log_decorator.log_info
     def cleanup(self):
         if self._ovs_port_info:
@@ -182,72 +128,6 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
         for table in self.driver.all_tables.values():
             self._ovs_flow_del(None, table)
 
-        if self.driver.proxy_arp:
-            self.log.info("Cleaning VRF network namespace %s", self.arp_netns)
-            # Detach network namespace veth pair device from OVS bridge
-            self._run_command(
-                "ovs-vsctl del-port %s %s" %
-                (self.bridge, get_ovsbr2arpns_if(self.arp_netns)),
-                run_as_root=True)
-            # Delete network namespace (deletes the veth pair as well)
-            self._run_command("ip netns delete %s" % self.arp_netns,
-                              run_as_root=True)
-
-    def _arp_net_ns_exists(self):
-        """Check if network namespace exist. """
-        (_, code) = self._run_command("ip netns pids %s" % self.arp_netns,
-                                      acceptable_return_codes=[0, 1],
-                                      run_as_root=True)
-        return code == 0
-
-    def _create_arp_netns_veth_pair(self, ovsbr_to_proxyarp_ns,
-                                    proxyarp_ns_to_ovsbr):
-        """Create a veth device pair, wtih one end in the ARP netns"""
-
-        try:
-            self._run_command("ip netns exec %s ip link del %s" %
-                              (self.arp_netns, proxyarp_ns_to_ovsbr),
-                              run_as_root=True,
-                              raise_on_error=False,
-                              acceptable_return_codes=[0, 1])
-            self._run_command("ip link del %s" % ovsbr_to_proxyarp_ns,
-                              run_as_root=True,
-                              raise_on_error=False,
-                              acceptable_return_codes=[0, 1])
-            self._run_command("ip link add %s mtu 65535 type veth peer name "
-                              "%s netns %s" % (ovsbr_to_proxyarp_ns,
-                                               proxyarp_ns_to_ovsbr,
-                                               self.arp_netns),
-                              run_as_root=True,
-                              acceptable_return_codes=[0, 2])
-            self._run_command("ip link set dev %s up" % ovsbr_to_proxyarp_ns,
-                              run_as_root=True)
-            self._run_command("ip netns exec %s ip link set dev %s up" %
-                              (self.arp_netns, proxyarp_ns_to_ovsbr),
-                              run_as_root=True)
-            self._run_command("ovs-vsctl del-port %s %s" %
-                              (self.bridge, ovsbr_to_proxyarp_ns),
-                              run_as_root=True,
-                              raise_on_error=False,
-                              acceptable_return_codes=[0, 1, 2])
-            self._run_command("ovs-vsctl add-port %s %s" %
-                              (self.bridge, ovsbr_to_proxyarp_ns),
-                              run_as_root=True)
-        except Exception:
-            self._run_command("ovs-vsctl del-port %s %s" %
-                              (self.bridge, ovsbr_to_proxyarp_ns),
-                              run_as_root=True,
-                              raise_on_error=False,
-                              acceptable_return_codes=[0, 1, 2])
-            self._run_command("ip netns exec %s ip link del %s" %
-                              (self.arp_netns, proxyarp_ns_to_ovsbr),
-                              run_as_root=True,
-                              raise_on_error=False)
-            self._run_command("ip link del %s" % ovsbr_to_proxyarp_ns,
-                              run_as_root=True,
-                              raise_on_error=False)
-            raise
-
     @log_decorator.log
     def _extract_mac_address(self, output):
         """Extract MAC address from command output"""
@@ -271,12 +151,18 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
             # succeeded even if the ping failed
 
         # Look in ARP cache to find remote MAC address
-        (output, _) = self._run_command("ip neigh show to %s" % (remote_ip))
+        (output, _) = self._run_command("ip neigh show to %s dev %s" %
+                                        (remote_ip, self.driver.bridge))
 
-        if not output or "FAILED" in output[0]:
+        if (not output or
+                "FAILED" in output[0] or
+                "INCOMPLETE" in output[0]):
             raise exc.RemotePEMACAddressNotFound(remote_ip)
 
-        return self._extract_mac_address(output[0])
+        try:
+            return self._extract_mac_address(output[0])
+        except Exception:
+            raise exc.RemotePEMACAddressNotFound(remote_ip)
 
     def _mtu_fixup(self, localport):
         # This is a hack, proper MTUs should actually be configured in the
@@ -376,6 +262,9 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
 
         return (port, localport_match, push_vlan_action, port_unplug_action)
 
+    def get_vlan_action(self):
+        return self.push_vlan_action if self.push_vlan_action else ""
+
     @log_decorator.log_info
     def update_fallback(self, fallback=None):
         if fallback:
@@ -384,10 +273,6 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
         if not self.fallback:
             return
 
-        vlan_action = ""
-        if self.push_vlan_action:
-            vlan_action = self.push_vlan_action
-
         for param in ('src_mac', 'dst_mac', 'ovs_port_number'):
             if not self.fallback.get(param):
                 self.log.error("fallback specified without '%s'", param)
@@ -399,10 +284,34 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
                            'mod_dl_src=%s,mod_dl_dst=%s,%soutput:%d' %
                            (self.fallback.get('src_mac'),
                             self.fallback.get('dst_mac'),
-                            vlan_action,
+                            self.get_vlan_action(),
                             self.fallback.get('ovs_port_number')),
                            self.driver.vrf_table,
-                           priority=DEFAULT_RULE_PRIORITY-1)
+                           priority=FALLBACK_PRIORITY)
+
+    @log_decorator.log_info
+    def setup_arp_responder(self, ovs_port):
+        actions = ARP_RESPONDER_ACTIONS % {
+            'mac': netaddr.EUI(GATEWAY_MAC, dialect=netaddr.mac_unix),
+            'vlan_action': self.get_vlan_action(),
+            'in_port': ovs_port
+        }
+
+        vrf_match = self._vrf_match(
+            'dl_dst=ff:ff:ff:ff:ff:ff,arp_op=0x1', proto='arp'
+        )
+        # Respond to all IP addresses if proxy ARP is enabled, otherwise only
+        # for gateway
+        if not self.driver.proxy_arp:
+            vrf_match += ',arp_tpa=%s' % self.gateway_ip
+
+        self._ovs_flow_add(vrf_match, actions, self.driver.vrf_table)
+
+    @log_decorator.log_info
+    def remove_arp_responder(self):
+        self._ovs_flow_del(
+            self._vrf_match(None, proto='arp'),
+            self.driver.vrf_table)
 
     def _check_vlan_use(self, push_vlan_action):
         # checks that if a vlan_action is used, it is the same
@@ -428,7 +337,8 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
             return (push_vlan_action, "strip_vlan,")
 
     @log_decorator.log
-    def vif_plugged(self, mac_address, ip_address, localport, label):
+    def vif_plugged(self, mac_address, ip_address, localport, label,
+                    direction):
 
         (ovs_port, localport_match, push_vlan_action, port_unplug_action) = (
             self._get_ovs_port_specifics(localport)
@@ -446,67 +356,54 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
         # Please consider this obsolete until it gets clean'd up
         self._mtu_fixup(localport)
 
-        # Map traffic from plugged port to VRF
+        if vpn_instance.forward_from_port(direction):
+            # Map traffic from plugged port to VRF
 
-        # Note that we first reset the in_port so that OVS will allow the
-        # packet to eventually go back to this port after a VRF lookup
-        # (the case where that happens is where in port is a patch-port on
-        # which different VLANs are used to reach different networks):
-        # patch port vlan X -- VRFX --- VRF Y -- patch-port vlan Y
-        #
-        # ( see http://docs.openvswitch.org/en/latest/faq/openflow/
-        # "Q: I added a flow to send packets out the ingress port..." )
-        actions = ('%s'
-                   'load:0->NXM_OF_IN_PORT[],'
-                   'set_field:%d->reg%d,'
-                   'resubmit(,%d)') % (
-            strip_vlan_action_str, self.instance_id, VRF_REGISTER,
-            self.driver.vrf_table
-        )
-        for proto in ('ip', 'arp'):
-            self._ovs_flow_add(join_s(localport_match, proto),
-                               actions,
-                               self.driver.input_table)
+            # Note that we first reset the in_port so that OVS will allow the
+            # packet to eventually go back to this port after a VRF lookup
+            # (the case where that happens is where in port is a patch-port on
+            # which different VLANs are used to reach different networks):
+            # patch port vlan X -- VRFX --- VRF Y -- patch-port vlan Y
+            #
+            # ( see http://docs.openvswitch.org/en/latest/faq/openflow/
+            # "Q: I added a flow to send packets out the ingress port..." )
+            actions = ('%s'
+                       'load:0->NXM_OF_IN_PORT[],'
+                       'set_field:%d->reg%d,'
+                       'resubmit(,%d)') % (
+                strip_vlan_action_str, self.instance_id, VRF_REGISTER,
+                self.driver.vrf_table
+            )
+            for proto in ('ip', 'arp'):
+                self._ovs_flow_add(join_s(localport_match, proto),
+                                   actions,
+                                   self.driver.input_table)
 
-        if self.driver.proxy_arp:
-            # Map ARP and IP traffic from ARP/GW netns to plugged port
+        # Map ARP responder if necessary
+        if not self._ovs_port_info:
+            self.setup_arp_responder(ovs_port)
 
-            # Map IP traffic from gateway to plugged port
-            self._ovs_flow_add(
-                join_s('in_port=%s' % self.arp_net_nsport,
-                       'ip',
-                       'nw_dst=%s' % ip_address),
-                '%soutput:%s' % (push_vlan_action_str, ovs_port),
-                self.driver.input_table)
+        if vpn_instance.forward_to_port(direction):
+            # Map incoming MPLS traffic going to the VM port
+            incoming_actions = ("%smod_dl_src:%s,mod_dl_dst:%s,output:%s" %
+                                (push_vlan_action_str, GATEWAY_MAC,
+                                 mac_address, ovs_port))
 
-            # ARP
-            self._ovs_flow_add(
-                join_s('in_port=%s' % self.arp_net_nsport,
-                       'arp',
-                       'dl_dst=%s' % mac_address),
-                '%soutput:%s' % (push_vlan_action_str, ovs_port),
-                self.driver.input_table)
-
-        # Map incoming MPLS traffic going to the VM port
-        incoming_actions = ("%smod_dl_src:%s,mod_dl_dst:%s,output:%s" %
-                            (push_vlan_action_str, GATEWAY_MAC,
-                             mac_address, ovs_port))
-
-        self._ovs_flow_add(self._match_mpls_in(label),
-                           "pop_mpls:0x0800,%s" % incoming_actions,
-                           self.driver.encap_in_table)
-
-        # additional incoming traffic rule for VXLAN
-        if self.driver.vxlan_encap:
-            self._ovs_flow_add(self._match_vxlan_in(label),
-                               incoming_actions,
+            self._ovs_flow_add(self._match_mpls_in(label),
+                               "pop_mpls:0x0800,%s" % incoming_actions,
                                self.driver.encap_in_table)
 
+            # additional incoming traffic rule for VXLAN
+            if self.driver.vxlan_encap:
+                self._ovs_flow_add(self._match_vxlan_in(label),
+                                   incoming_actions,
+                                   self.driver.encap_in_table)
+
         # Add OVS port number in list for local port plugged in VRF
         # FIXME: check check check, is linuxif the right key??
-        self.log.debug("Adding OVS port %s with numbers (%s,%s) for address "
-                       "%s to ports plugged in VRF list",
-                       localport['linuxif'], ovs_port, ovs_port, ip_address)
+        self.log.debug("Adding OVS port %s with port %s for address "
+                       "%s, to the list of ports plugged in VRF",
+                       localport['linuxif'], ovs_port, ip_address)
         self._ovs_port_info[localport['linuxif']] = {
             "localport_match": localport_match,
             "port_unplug_action": port_unplug_action,
@@ -521,44 +418,33 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
 
     @log_decorator.log
     def vif_unplugged(self, mac_address, ip_address, localport, label,
-                      last_endpoint=True):
+                      direction, last_endpoint=True):
 
         localport_match = self._ovs_port_info[
             localport['linuxif']]['localport_match']
         port_unplug_action = self._ovs_port_info[
             localport['linuxif']]['port_unplug_action']
 
-        # Unmap incoming MPLS traffic going to the VM port
-        self._ovs_flow_del(self._match_mpls_in(label),
-                           self.driver.encap_in_table)
-
-        # Unmap incoming VXLAN traffic...
-        if self.driver.vxlan_encap:
-            self._ovs_flow_del(self._match_vxlan_in(label),
+        if vpn_instance.forward_to_port(direction):
+            # Unmap incoming MPLS traffic going to the VM port
+            self._ovs_flow_del(self._match_mpls_in(label),
                                self.driver.encap_in_table)
 
-        # Unmap all traffic from plugged port
-        if last_endpoint:
-            self._ovs_flow_del(localport_match, self.driver.input_table)
-
-        if self.driver.proxy_arp:
-            # Unmap IP traffic from ARP/GW netns to plugged port
-            self._ovs_flow_del(
-                join_s('in_port=%s' % self.arp_net_nsport,
-                       'ip',
-                       'nw_dst=%s' % ip_address),
-                self.driver.input_table)
-
-            # Unmap ARP traffic from ARP/GW netns to plugged port
-            self._ovs_flow_del(
-                join_s('in_port=%s' % self.arp_net_nsport,
-                       'arp',
-                       'dl_dst=%s' % mac_address),
-                self.driver.input_table)
+            # Unmap incoming VXLAN traffic...
+            if self.driver.vxlan_encap:
+                self._ovs_flow_del(self._match_vxlan_in(label),
+                                   self.driver.encap_in_table)
 
         if last_endpoint:
+            if vpn_instance.forward_from_port(direction):
+                # Unmap all traffic from plugged port
+                self._ovs_flow_del(localport_match, self.driver.input_table)
+
+            # Unmap ARP responder
+            self.remove_arp_responder()
+
+            # Run port unplug action if necessary (OVS port delete)
             if port_unplug_action:
-                # Run port unplug action if necessary (OVS port delete)
                 self._run_command(port_unplug_action,
                                   run_as_root=True,
                                   acceptable_return_codes=[0, 1])
@@ -616,11 +502,6 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
                     self.driver.mpls_if_mac_address, remote_pe_mac_address,
                     self.driver.ovs_mpls_if_port_number)
 
-    def _match_default_route_prefix(self, prefix):
-        return ('nw_dst=%s' % prefix
-                if netaddr.IPNetwork(prefix).prefixlen != 0
-                else None)
-
     def _cookie(self, add=False):
         mask = ""
         if not add:
@@ -635,8 +516,10 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
         return self._vrf_match(join_s('reg%d=%d' % (LB_HOP_REGISTER, index),
                                       match))
 
-    def _get_lb_flows_to_add(self, prefix, nw_dst_match):
+    def _get_lb_flows_to_add(self, prefix):
         dec_ttl_action = ""
+
+        # if destination in same subnet as the VRF, don't decrement TTL
         if netaddr.IPNetwork(prefix) not in netaddr.IPNetwork("%s/%s" %
                                                               (self.gateway_ip,
                                                                self.mask)):
@@ -650,32 +533,30 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
                                                       endpoint['encaps'])
 
             lb_endpoint_flow = self._ovs_flow_add(
-                self._vrf_lb_match(index, nw_dst_match),
+                self._vrf_lb_match(index, _match_from_prefix(prefix)),
                 join_s(dec_ttl_action,
                        label_action,
                        output_action),
                 self.driver.post_hash_vrf_table,
+                priority=_priority_from_prefix(prefix),
                 return_flow=True)
 
             flows_to_add.append(('add', lb_endpoint_flow))
 
         return flows_to_add
 
-    def _get_lb_flows_to_del(self, prefix, nw_dst_match):
+    def _get_lb_flows_to_del(self, prefix):
         flows_to_del = []
         for index, _ in enumerate(self._lb_endpoints[prefix]):
             lb_endpoint_flow = self._ovs_flow_del(
-                self._vrf_lb_match(index, nw_dst_match),
+                self._vrf_lb_match(index, _match_from_prefix(prefix)),
                 self.driver.post_hash_vrf_table,
                 return_flow=True)
             flows_to_del.append(('del', lb_endpoint_flow))
 
         return flows_to_del
 
-    def _get_lb_multipath_flow_mod(self, prefix, nw_dst_match):
-        self.log.debug('Prefix %s: nw_dst_match "%s", %s', prefix,
-                       nw_dst_match, self._lb_endpoints[prefix])
-
+    def _get_lb_multipath_flow_mod(self, prefix):
         if self._lb_endpoints[prefix]:
             multipath_action = ('multipath(symmetric_l3l4+udp,1024,hrw,%d,0,'
                                 'NXM_NX_REG%d[])' % (
@@ -686,22 +567,27 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
                 'resubmit(,%d)' % self.driver.post_hash_vrf_table)
 
             lb_multipath_flow = self._ovs_flow_add(
-                self._vrf_match(nw_dst_match),
+                self._vrf_match(_match_from_prefix(prefix)),
                 join_s(multipath_action, multipath_output),
                 self.driver.vrf_table,
+                priority=_priority_from_prefix(prefix),
                 return_flow=True)
             self.log.debug('Multipath flow: %s', lb_multipath_flow)
             if len(self._lb_endpoints[prefix]) > 1:
+                # TODO(tmorin): should use consts here from some OVS lib
                 return 'modify_strict', lb_multipath_flow
             else:
+                # TODO(tmorin): should use consts here from some OVS lib
                 return 'add', lb_multipath_flow
         else:
             lb_multipath_flow = (
-                self._ovs_flow_del(self._vrf_match(nw_dst_match),
+                self._ovs_flow_del(self._vrf_match(_match_from_prefix(prefix)),
                                    self.driver.vrf_table,
+                                   priority=_priority_from_prefix(prefix),
                                    strict=True,
                                    return_flow=True)
             )
+            # TODO(tmorin): should use consts here from some OVS lib
             return 'delete_strict', lb_multipath_flow
 
     @log_decorator.log_info
@@ -726,14 +612,13 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
 
         if (prefix in self._lb_endpoints and
                 lb_endpoint_info in self._lb_endpoints[prefix]):
+            self.log.debug("Dataplane already in place for %s, %s, skipping",
+                           prefix, lb_endpoint_info)
             return
 
-        # Check if prefix is a default route
-        nw_dst_match = self._match_default_route_prefix(prefix)
-
         lb_flows = list()
         if prefix in self._lb_endpoints:
-            lb_flows.extend(self._get_lb_flows_to_del(prefix, nw_dst_match))
+            lb_flows.extend(self._get_lb_flows_to_del(prefix))
         else:
             self._lb_endpoints[prefix] = list()
 
@@ -742,12 +627,12 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
         if len(self._lb_endpoints[prefix]) > 1:
             self._lb_endpoints[prefix] = sorted(
                 self._lb_endpoints[prefix],
-                key=lambda endpoint: endpoint['lb_consistent_hash_order']
+                key=operator.itemgetter('lb_consistent_hash_order')
             )
 
-        lb_flows.append(self._get_lb_multipath_flow_mod(prefix, nw_dst_match))
+        lb_flows.append(self._get_lb_multipath_flow_mod(prefix))
 
-        lb_flows.extend(self._get_lb_flows_to_add(prefix, nw_dst_match))
+        lb_flows.extend(self._get_lb_flows_to_add(prefix))
 
         self.driver._ovs_flow_mods(lb_flows)
 
@@ -755,13 +640,10 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
     def remove_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
                                              nlri, encaps,
                                              lb_consistent_hash_order=0):
-        # Check if prefix is a default route
-        nw_dst_match = self._match_default_route_prefix(prefix)
-
         if prefix in self._lb_endpoints:
             lb_flows = []
 
-            lb_flows.extend(self._get_lb_flows_to_del(prefix, nw_dst_match))
+            lb_flows.extend(self._get_lb_flows_to_del(prefix))
 
             self._lb_endpoints[prefix].remove(
                 {'label': label,
@@ -770,16 +652,18 @@ class MPLSOVSVRFDataplane(dp_drivers.VPNInstanceDataplane):
                  'lb_consistent_hash_order': lb_consistent_hash_order}
             )
 
-            lb_flows.append(self._get_lb_multipath_flow_mod(prefix,
-                                                            nw_dst_match))
+            lb_flows.append(self._get_lb_multipath_flow_mod(prefix))
 
             if self._lb_endpoints[prefix]:
-                lb_flows.extend(self._get_lb_flows_to_add(prefix,
-                                                          nw_dst_match))
+                lb_flows.extend(self._get_lb_flows_to_add(prefix))
             else:
                 del self._lb_endpoints[prefix]
 
             self.driver._ovs_flow_mods(lb_flows)
+        else:
+            self.log.warning("remove_dataplane_for_remote_endpoint called, "
+                             "for %s, but we don't know about this prefix",
+                             prefix)
 
     def _create_flow_match_from_tc(self, classifier):
         flow_match = ''
@@ -895,7 +779,7 @@ class MPLSOVSDataplaneDriver(dp_drivers.DataplaneDriver):
     dataplane_instance_class = MPLSOVSVRFDataplane
     type = consts.IPVPN
     ecmp_support = True
-    required_ovs_version = "2.5.0"
+    required_ovs_version = "2.8.0"
 
     driver_opts = [
         cfg.StrOpt("mpls_interface",
@@ -905,20 +789,22 @@ class MPLSOVSDataplaneDriver(dp_drivers.DataplaneDriver):
         cfg.StrOpt("mpls_over_gre",
                    choices=['auto', 'True', 'False'],
                    default="auto",
+                   advanced=True,
                    help=("Force the use of MPLS/GRE even with "
                          "mpls_interface specified")),
-        cfg.BoolOpt("proxy_arp", default=True,
-                    help=("A netns will be connected to each VRF, will "
-                          "be setup with the gateway IP address, and will "
-                          "reply to all ARP requests")),
+        cfg.BoolOpt("proxy_arp", default=False,
+                    advanced=True,
+                    help=("Activate ARP responder per VRF for all IP "
+                          "addresses (only for gateway IP by default)")),
         cfg.BoolOpt("vxlan_encap", default=False,
+                    advanced=True,
                     help=("Be ready to receive VPN traffic as VXLAN, and to "
                           "preferrably send traffic as VXLAN when advertised "
                           "by the remote end")),
-        cfg.StrOpt("ovs_bridge", default="br-mpls"),
-        cfg.IntOpt("input_table", default=0),
-        cfg.IntOpt("ovs_table_id_start", default=1),
-        cfg.StrOpt("gre_tunnel", default="mpls_gre",
+        cfg.StrOpt("ovs_bridge", default="br-mpls", advanced=True),
+        cfg.IntOpt("input_table", default=0, advanced=True),
+        cfg.IntOpt("ovs_table_id_start", default=1, advanced=True),
+        cfg.StrOpt("gre_tunnel", default="mpls_gre", advanced=True,
                    help="OVS interface name for MPLS/GRE encap"),
         cfg.ListOpt("gre_tunnel_options", default=[],
                     item_type=types.String(),
@@ -927,7 +813,7 @@ class MPLSOVSDataplaneDriver(dp_drivers.DataplaneDriver):
                           ", ...') that will be added as OVS tunnel "
                           "interface options (e.g. 'options:packet_type="
                           "legacy_l3 options:...')")),
-        cfg.IntOpt("ovsbr_interfaces_mtu")
+        cfg.IntOpt("ovsbr_interfaces_mtu", advanced=True)
     ]
 
     def __init__(self):
@@ -1108,38 +994,6 @@ class MPLSOVSDataplaneDriver(dp_drivers.DataplaneDriver):
             self.log.info("No OVS bridge (%s), no need to cleanup OVS rules",
                           self.bridge)
 
-        # Flush network namespaces for ARP responders and
-        # corresponding veth pair devices
-        cmd = r"ip netns list | cut '-d ' -f 1 | grep -v '\<q' | grep '%s'"
-        (output, _) = self._run_command(cmd % ARPNETNS_PREFIX,
-                                        raise_on_error=False,
-                                        acceptable_return_codes=[0, 1])
-        if not output:
-            self.log.debug("No network namespaces configured")
-        else:
-            for namespace_id in output:
-                self.log.info("Cleaning up netns %s", namespace_id)
-                self._run_command("ip netns delete %s" % namespace_id,
-                                  run_as_root=True,
-                                  raise_on_error=False)
-                self._run_command(
-                    "ovs-vsctl del-port %s %s" % (
-                        self.bridge,
-                        get_ovsbr2arpns_if(namespace_id)),
-                    run_as_root=True,
-                    acceptable_return_codes=[0, 1, 2],
-                    raise_on_error=False)
-            if self.log.debug:
-                self.log.debug("All network namespaces have been flushed")
-                self._run_command("ip netns")
-
-                self.log.debug("All network namespace veth pairs flushed")
-                self._run_command("ifconfig")
-                self._run_command("ovs-vsctl list-ports %s" % self.bridge,
-                                  run_as_root=True,
-                                  acceptable_return_codes=[0, 1])
-
-    @log_decorator.log_info
     def initialize(self):
         if self.use_gre:
             self.log.info("Setting up tunnel for MPLS/GRE (%s)",
@@ -1200,6 +1054,10 @@ class MPLSOVSDataplaneDriver(dp_drivers.DataplaneDriver):
                                "resubmit(,%d)" % self.encap_in_table,
                                self.input_table)
 
+    def validate_directions(self, direction):
+        # this driver supports all combinations of directions
+        pass
+
     def find_ovs_port(self, dev_name):
         """Find OVS port number from port name"""
 
@@ -1284,7 +1142,7 @@ class MPLSOVSDataplaneDriver(dp_drivers.DataplaneDriver):
         for (table_name, table_id) in self.all_tables.items():
             output.update({
                 "%s (%d)" % (table_name, table_id): self._run_command(
-                    "ovs-ofctl dump-flows %s '%s' %s" % (
+                    "ovs-ofctl dump-flows --names %s '%s' %s" % (
                         self.bridge,
                         join_s("table=%d" % table_id, cookie_spec),
                         OVS_DUMP_FLOW_FILTER),
