@@ -27,38 +27,27 @@ from networking_bagpipe.bagpipe_bgp.common import run_command
 from networking_bagpipe.bagpipe_bgp.common import utils
 from networking_bagpipe.bagpipe_bgp import constants
 from networking_bagpipe.bagpipe_bgp.engine import bgp_manager
-from networking_bagpipe.bagpipe_bgp.engine import exa
 from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
 from networking_bagpipe.bagpipe_bgp.vpn import evpn
 from networking_bagpipe.bagpipe_bgp.vpn import ipvpn
 from networking_bagpipe.bagpipe_bgp.vpn import label_allocator
 from networking_bagpipe.bagpipe_bgp.vpn import rd_allocator
 
+from neutron_lib import exceptions
+
 
 LOG = logging.getLogger(__name__)
 
+INSTANCE_ID_MAX = 2**32-1
+
 
 def redirect_instance_extid(instance_type, rt):
     '''generate the ext intance id of a redirection VPN instance'''
     return "redirect-to-%s-%s" % (instance_type, rt.replace(":", "_"))
 
 
-def convert_route_targets(orig_list):
-    assert isinstance(orig_list, list)
-    list_ = []
-    for rt in orig_list:
-        if rt == '':
-            continue
-        try:
-            asn, nn = rt.split(':')
-            list_.append(exa.RouteTarget(int(asn), int(nn)))
-        except Exception:
-            raise Exception("Malformed route target: '%s'" % rt)
-    return list_
-
-
-class NoSuchVPNInstance(Exception):
-    pass
+class MaxInstanceIDReached(exceptions.NeutronException):
+    _message = "Could not create VPN instance: max instance id was reached"
 
 
 class VPNManager(lg.LookingGlassMixin):
@@ -257,6 +246,9 @@ class VPNManager(lg.LookingGlassMixin):
 
         # unique internal vpn instance id
         instance_id = self.next_vpn_instance_id
+        if instance_id > INSTANCE_ID_MAX:
+            raise MaxInstanceIDReached()
+
         self.next_vpn_instance_id += 1
 
         vpn_instance = vpn_instance_class(self, dataplane_driver,
@@ -311,18 +303,19 @@ class VPNManager(lg.LookingGlassMixin):
         readvertise = params.get('readvertise')
         attract_traffic = params.get('attract_traffic')
         lb_consistent_hash_order = params.get('lb_consistent_hash_order')
+        local_pref = params.get('local_pref')
         fallback = params.get('fallback')
         vni = params.get('vni')
 
         ip_address_prefix = params.get('ip_address_prefix')
 
         # Convert route target string to RouteTarget dictionary
-        import_rts = convert_route_targets(import_rts)
-        export_rts = convert_route_targets(export_rts)
+        import_rts = utils.convert_route_targets(import_rts)
+        export_rts = utils.convert_route_targets(export_rts)
 
         if readvertise:
             try:
-                readvertise = {k: convert_route_targets(readvertise[k])
+                readvertise = {k: utils.convert_route_targets(readvertise[k])
                                for k in ['from_rt', 'to_rt']}
             except KeyError as e:
                 raise Exception("Wrong 'readvertise' parameters: %s" % e)
@@ -330,7 +323,9 @@ class VPNManager(lg.LookingGlassMixin):
         if attract_traffic:
             try:
                 attract_traffic['redirect_rts'] = (
-                    convert_route_targets(attract_traffic['redirect_rts']))
+                    utils.convert_route_targets(
+                        attract_traffic['redirect_rts'])
+                )
             except KeyError as e:
                 raise Exception("Wrong 'attract_traffic' parameters: %s" % e)
 
@@ -347,6 +342,8 @@ class VPNManager(lg.LookingGlassMixin):
             external_instance_id, instance_type, import_rts, export_rts,
             gateway_ip, mask, readvertise, attract_traffic, fallback, **kwargs)
 
+        vpn_instance.description = params.get('instance_description')
+
         # Check if new route target import/export must be updated
         if not ((set(vpn_instance.import_rts) == set(import_rts)) and
                 (set(vpn_instance.export_rts) == set(export_rts))):
@@ -360,9 +357,14 @@ class VPNManager(lg.LookingGlassMixin):
             # to create, connected to an existing evpn instance
             self._attach_evpn_2_ipvpn(localport, vpn_instance)
 
+        plug_kwargs = {}
+        plug_kwargs['description'] = params.get('description')
+        plug_kwargs['direction'] = params.get('direction')
+
         # Plug VIF to VPN instance
         vpn_instance.vif_plugged(mac_address, ip_address_prefix, localport,
-                                 advertise_subnet, lb_consistent_hash_order)
+                                 advertise_subnet, lb_consistent_hash_order,
+                                 local_pref, **plug_kwargs)
 
     @log_decorator.log_info
     def unplug_vif_from_vpn(self, **params):
@@ -375,8 +377,6 @@ class VPNManager(lg.LookingGlassMixin):
         external_instance_id = params.get('external_instance_id')
         mac_address = params.get('mac_address')
         localport = params.get('localport')
-        readvertise = params.get('readvertise')
-
         ip_address_prefix = params.get('ip_address_prefix')
 
         # Retrieve VPN instance or raise exception if does not exist
@@ -388,7 +388,7 @@ class VPNManager(lg.LookingGlassMixin):
             raise exc.VPNNotFound(external_instance_id)
 
         # Unplug VIF from VPN instance
-        vpn_instance.vif_unplugged(mac_address, ip_address_prefix, readvertise)
+        vpn_instance.vif_unplugged(mac_address, ip_address_prefix)
 
         if vpn_instance.type == constants.IPVPN and 'evpn' in localport:
             self._detach_evpn_2_ipvpn(vpn_instance)
@@ -404,7 +404,7 @@ class VPNManager(lg.LookingGlassMixin):
                  "target %s", external_instance_id, redirect_rt)
 
         # Convert route target string to RouteTarget dictionary
-        import_rts = convert_route_targets([redirect_rt])
+        import_rts = utils.convert_route_targets([redirect_rt])
 
         # Retrieve a redirect VPN instance or create a new one if none exists
         # yet
@@ -498,10 +498,8 @@ class VPNManager(lg.LookingGlassMixin):
         }
 
     def get_lg_vpn_list(self):
-        return [{"id": id,
-                 "name": instance.name}
-                for (id, instance)
-                in six.iteritems(self.vpn_instances)]
+        return [instance.get_lg_summary()
+                for instance in list(self.vpn_instances.values())]
 
     def get_lg_vpn_from_path_item(self, path_item):
         return self.vpn_instances[path_item]
