@@ -30,14 +30,24 @@ from networking_bagpipe.bagpipe_bgp.common import exceptions as exc
 from networking_bagpipe.bagpipe_bgp.common import log_decorator
 from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
 from networking_bagpipe.bagpipe_bgp.common import utils
+from networking_bagpipe.bagpipe_bgp import constants
 from networking_bagpipe.bagpipe_bgp import engine
 from networking_bagpipe.bagpipe_bgp.engine import exa
 from networking_bagpipe.bagpipe_bgp.engine import flowspec
 from networking_bagpipe.bagpipe_bgp.engine import tracker_worker
 
-
 LOG = logging.getLogger(__name__)
 
+DEFAULT_LOCAL_PREF = 100
+
+
+def forward_to_port(direction):
+    return direction in (None, constants.BOTH, constants.TO_PORT)
+
+
+def forward_from_port(direction):
+    return direction in (None, constants.BOTH, constants.FROM_PORT)
+
 
 class TrafficClassifier(object):
 
@@ -246,6 +256,8 @@ class VPNInstance(tracker_worker.TrackerWorker,
         self.instance_type = self.__class__.__name__
         self.instance_id = instance_id
 
+        self.description = None
+
         threading.Thread.__init__(self)
         self.setDaemon(True)
 
@@ -263,7 +275,7 @@ class VPNInstance(tracker_worker.TrackerWorker,
         lg.LookingGlassLocalLogger.__init__(self,
                                             "%s-%d" % (self.instance_type,
                                                        self.instance_id))
-        self.lock = threading.Lock()
+        self.lock = threading.RLock()
 
         self.import_rts = import_rts
         self.export_rts = export_rts
@@ -293,14 +305,25 @@ class VPNInstance(tracker_worker.TrackerWorker,
 
         self.localport_data = dict()
 
-        # One local port -> List of endpoints (MAC and IP addresses tuple)
-        self.localport_2_endpoints = dict()
+        # One local port -> set of endpoints (MAC and IP addresses tuple)
+        self.localport_2_endpoints = collections.defaultdict(set)
         # One endpoint (MAC and IP addresses tuple) -> One route distinguisher
         self.endpoint_2_rd = dict()
+        # endpoint (MAC and IP addresses tuple) -> route entry
+        self.endpoint_2_route = dict()
         # One MAC address -> One local port
         self.mac_2_localport_data = dict()
         # One IP address ->  Multiple MAC address
-        self.ip_address_2_mac = dict()
+        self.ip_address_2_mac = collections.defaultdict(set)
+
+        # One endpoint (MAC and IP addresses tuple) -> BGP local_pref
+        self.endpoint_2_lp = dict()
+
+        # endpoint (MAC and IP addresses tuple) -> description
+        self.endpoint_2_desc = dict()
+
+        # endpoint (MAC and IP addresses tuple) -> direction
+        self.endpoint_2_direction = dict()
 
         # Redirected instances list from which traffic is attracted (based on
         # FlowSpec 5-tuple classification)
@@ -332,11 +355,40 @@ class VPNInstance(tracker_worker.TrackerWorker,
             self.log.debug("readvertise not enabled")
             self.readvertise = False
 
-        if self.readvertise and attract_traffic:
+        self.attract_static_dest_prefixes = None
+
+        if (attract_traffic and (self.readvertise or
+                                 attract_traffic.get('to_rt'))):
+            # Convert route target string to RouteTarget dictionary
+            attract_to_rts = attract_traffic.get('to_rt')
+
+            if attract_to_rts:
+                converted_to_rts = (
+                    utils.convert_route_targets(attract_to_rts)
+                )
+                if self.readvertise:
+                    if converted_to_rts != self.readvertise_to_rts:
+                        raise exc.APIException("if both are set, then "
+                                               "'attract_traffic/to_rt' and "
+                                               "'readvertise/to_rt' have to "
+                                               "be equal")
+                else:
+                    self.readvertise_to_rts = converted_to_rts
+
+                try:
+                    self.attract_static_dest_prefixes = (
+                        attract_traffic['static_destination_prefixes']
+                    )
+                except KeyError:
+                    raise exc.APIException("'attract_traffic/to_rt' specified "
+                                           "without "
+                                           "'static_destination_prefixes'")
+
             if len(self.readvertise_to_rts) != 1:
                 raise exc.APIException("attract_traffic requires exactly one "
                                        "RT to be provided in "
-                                       "readvertise/to_rt")
+                                       "'readvertise/to_rt' and "
+                                       "'attract_traffic/to_rt'")
             self.attract_traffic = True
             self.attract_rts = attract_traffic['redirect_rts']
             try:
@@ -354,11 +406,14 @@ class VPNInstance(tracker_worker.TrackerWorker,
         self.dataplane.update_fallback(fallback)
 
     @utils.synchronized
-    def stop(self):
-        self._stop()
-
     @log_decorator.log
-    def _stop(self):
+    def stop(self):
+        self.stop_event_loop()
+
+        if self.dataplane.needs_cleanup_assist():
+            self.log.debug("Dataplane driver needs cleanup assistance")
+            self.synthesize_withdraw_all()
+
         self.dataplane.cleanup()
         if not self.forced_vni:
             self.manager.label_allocator.release(self.instance_label)
@@ -372,7 +427,7 @@ class VPNInstance(tracker_worker.TrackerWorker,
     def stop_if_empty(self):
         self.log.debug("localport_2_endpoints: %s", self.localport_2_endpoints)
         if self.is_empty():
-            self._stop()
+            self.stop()
             return True
 
         return False
@@ -397,44 +452,45 @@ class VPNInstance(tracker_worker.TrackerWorker,
         self.log.debug("%s %d - Removed Import RTs: %s",
                        self.instance_type, self.instance_id, removed_import_rt)
 
-        # Register to BGP with these route targets
-        for rt in added_import_rt:
-            self._subscribe(self.afi, self.safi, rt)
-            self._subscribe(self.afi, exa.SAFI.flow_vpn, rt)
-
         # Unregister from BGP with these route targets
         for rt in removed_import_rt:
             self._unsubscribe(self.afi, self.safi, rt)
             self._unsubscribe(self.afi, exa.SAFI.flow_vpn, rt)
 
         # Update import and export route targets
+        # (needs to be done before subscribe or we get a race where
+        # VRF._imported rejects route that it's supposed to use)
         self.import_rts = new_import_rts
 
+        # Register to BGP with these route targets
+        for rt in added_import_rt:
+            self._subscribe(self.afi, self.safi, rt)
+            self._subscribe(self.afi, exa.SAFI.flow_vpn, rt)
+
         # Re-advertise all routes with new export RTs
         self.log.debug("Exports RTs: %s -> %s", self.export_rts,
                        new_export_rts)
         if frozenset(new_export_rts) != frozenset(self.export_rts):
             self.log.debug("Will re-export routes with new RTs")
             self.export_rts = new_export_rts
-            # FIXME: we should only update the routes that
-            # are routes of ports plugged to the VPN instance,
-            # not all routes which would wrongly include
-            # routes that we re-advertise between RTs
-            for route_entry in self.get_route_entries():
-                self.log.info("Re-advertising route %s with updated RTs (%s)",
-                              route_entry.nlri, new_export_rts)
+            for route_entry in self.endpoint_2_route.values():
+                if new_export_rts:
+                    self.log.info("Re-advertising route %s with updated RTs "
+                                  "(%s)", route_entry.nlri, new_export_rts)
 
-                updated_route_entry = engine.RouteEntry(
-                    route_entry.nlri,
-                    None,
-                    copy.copy(route_entry.attributes))
-                # reset the route_targets
-                # will RTs originally present in route_entry.attributes
-                updated_route_entry.set_route_targets(self.export_rts)
+                    updated_route_entry = engine.RouteEntry(
+                        route_entry.nlri,
+                        None,
+                        copy.copy(route_entry.attributes))
+                    # reset the route_targets, overwriting RTs originally
+                    # present in route_entry.attributes
+                    updated_route_entry.set_route_targets(self.export_rts)
 
-                self.log.debug("   updated route: %s", updated_route_entry)
+                    self.log.debug("   updated route: %s", updated_route_entry)
 
-                self._advertise_route(updated_route_entry)
+                    self._advertise_route(updated_route_entry)
+                else:
+                    self._withdraw_route(route_entry)
 
     def update_fallback(self, fallback):
         if self.fallback != fallback and fallback is not None:
@@ -474,7 +530,8 @@ class VPNInstance(tracker_worker.TrackerWorker,
 
     def synthesize_vif_bgp_route(self, mac_address, ip_prefix, plen, label,
                                  lb_consistent_hash_order,
-                                 route_distinguisher=None):
+                                 route_distinguisher=None,
+                                 local_pref=None):
         rd = route_distinguisher if route_distinguisher else self.instance_rd
         route_entry = self.generate_vif_bgp_route(mac_address, ip_prefix, plen,
                                                   label, rd)
@@ -488,6 +545,9 @@ class VPNInstance(tracker_worker.TrackerWorker,
             exa.ConsistentHashSortOrder(lb_consistent_hash_order))
         route_entry.attributes.add(ecommunities)
 
+        route_entry.attributes.add(exa.LocalPreference(local_pref
+                                                       or DEFAULT_LOCAL_PREF))
+
         self.log.debug("Synthesized Vif route entry: %s", route_entry)
         return route_entry
 
@@ -531,8 +591,11 @@ class VPNInstance(tracker_worker.TrackerWorker,
                 isinstance(params['local_port'], six.text_type)):
             params['local_port'] = {'linuxif': params['local_port']}
 
-        # if import_rt or export_rt are strings, convert them into lists
         for param in ('import_rt', 'export_rt'):
+            if param not in params:
+                continue
+
+            # if import_rt or export_rt are strings, convert them into lists
             if (isinstance(params[param], six.string_types) or
                     isinstance(params[param], six.text_type)):
                 try:
@@ -540,6 +603,8 @@ class VPNInstance(tracker_worker.TrackerWorker,
                 except Exception:
                     raise exc.APIException("Unable to parse RT string into "
                                            " a list: '%s'" % params[param])
+            # remove duplicates
+            params[param] = list(set(params[param]))
 
         if not ('linuxif' in params['local_port'] or
                 'evpn' in params['local_port']):
@@ -555,6 +620,9 @@ class VPNInstance(tracker_worker.TrackerWorker,
         else:
             raise exc.MalformedIPAddress(params['ip_address'])
 
+        # NOTE(tmorin): improve the IP check above and validate MAC address
+        # format
+
         if not isinstance(params.get('advertise_subnet', False), bool):
             raise exc.APIException("'advertise_subnet' must be a boolean")
 
@@ -585,6 +653,11 @@ class VPNInstance(tracker_worker.TrackerWorker,
             'lb_consistent_hash_order', 0)
         params['vni'] = params.get('vni', 0)
 
+        params['direction'] = params.get('direction') or constants.BOTH
+        if params['direction'] not in constants.ALL_DIRECTIONS:
+            raise exc.APIException("direction should be one of: %s" %
+                                   ', '.join(constants.ALL_DIRECTIONS))
+
         cls.translate_api_internal(params)
 
     @classmethod
@@ -592,12 +665,21 @@ class VPNInstance(tracker_worker.TrackerWorker,
         cls.validate_convert_params(params)
         cls.translate_api_internal(params)
 
+    def _rd_for_endpoint(self, endpoint, message):
+        # endpoint is a (mac, ip_address_prefix) tuple
+        rd = self.endpoint_2_rd.get(endpoint)
+        if not rd:
+            rd = self.manager.rd_allocator.get_new_rd(message)
+            self.endpoint_2_rd[endpoint] = rd
+        return rd
+
     @utils.synchronized
     @log_decorator.log_info
     def vif_plugged(self, mac_address, ip_address_prefix, localport,
                     advertise_subnet=False,
-                    lb_consistent_hash_order=0):
+                    lb_consistent_hash_order=0, local_pref=None, **kwargs):
         linuxif = localport['linuxif']
+        endpoint = (mac_address, ip_address_prefix)
         # Check if this port has already been plugged
         # - Verify port informations consistency
         if mac_address in self.mac_2_localport_data:
@@ -614,6 +696,8 @@ class VPNInstance(tracker_worker.TrackerWorker,
                                         localport))
 
         try:
+            self.endpoint_2_desc[endpoint] = kwargs.get('description')
+
             # Parse address/mask
             (ip_prefix, plen) = self._parse_ipaddress_prefix(ip_address_prefix)
 
@@ -622,6 +706,7 @@ class VPNInstance(tracker_worker.TrackerWorker,
                 plen = 32
 
             # - Verify (MAC address, IP address) tuple consistency
+            refresh_only = False
             if ip_address_prefix in self.ip_address_2_mac and plen == 32:
                 if mac_address not in self.ip_address_2_mac[ip_address_prefix]:
                     raise exc.APIException("Inconsistent endpoint info: %s "
@@ -629,6 +714,8 @@ class VPNInstance(tracker_worker.TrackerWorker,
                                            "different from %s" %
                                            (ip_address_prefix, mac_address))
                 else:
+                    LOG.debug("IP/MAC already plugged, only updating route")
+                    refresh_only = True
                     return
 
             self.log.debug("Plugging port (%s)", ip_prefix)
@@ -644,91 +731,86 @@ class VPNInstance(tracker_worker.TrackerWorker,
                 pdata["port_info"] = localport
                 pdata["lb_consistent_hash_order"] = lb_consistent_hash_order
 
-            endpoint_rd = self.manager.rd_allocator.get_new_rd(
-                "Route distinguisher for %s %d, interface %s, "
-                "endpoint %s/%s" % (self.instance_type, self.instance_id,
-                                    linuxif, mac_address,
-                                    ip_address_prefix)
-            )
+            direction = kwargs.get('direction')
 
-            rd = self.instance_rd if plen == 32 else endpoint_rd
+            if not refresh_only:
+                self.dp_driver.validate_directions(direction)
+                # Call driver to setup the dataplane for incoming traffic
+                self.dataplane.vif_plugged(mac_address, ip_prefix,
+                                           localport, pdata['label'],
+                                           direction)
 
-            # Call driver to setup the dataplane for incoming traffic
-            self.dataplane.vif_plugged(mac_address, ip_prefix,
-                                       localport, pdata['label'])
+            if forward_to_port(direction):
+                endpoint_rd = self._rd_for_endpoint(
+                    endpoint,
+                    "Route distinguisher for %s %d, interface %s, "
+                    "endpoint %s/%s" % (self.instance_type, self.instance_id,
+                                        linuxif, mac_address,
+                                        ip_address_prefix)
+                )
 
-            self.log.info("Synthesizing and advertising BGP route for VIF %s "
-                          "endpoint (%s, %s/%d)", linuxif,
-                          mac_address, ip_prefix, plen)
-            route_entry = self.synthesize_vif_bgp_route(
-                mac_address, ip_prefix, plen,
-                pdata['label'], lb_consistent_hash_order, rd)
+                rd = self.instance_rd if plen == 32 else endpoint_rd
 
-            self._advertise_route(route_entry)
+                self.log.info("Synthesizing and advertising BGP route for VIF "
+                              "%s endpoint (%s, %s/%d)", linuxif,
+                              mac_address, ip_prefix, plen)
+                route_entry = self.synthesize_vif_bgp_route(
+                    mac_address, ip_prefix, plen,
+                    pdata['label'], lb_consistent_hash_order, rd, local_pref)
 
-            if linuxif not in self.localport_2_endpoints:
-                self.localport_2_endpoints[linuxif] = list()
+                self._advertise_route(route_entry)
+                self.endpoint_2_route[endpoint] = route_entry
+                self.endpoint_2_rd[endpoint] = endpoint_rd
 
-            self.localport_2_endpoints[linuxif].append(
-                {'mac': mac_address, 'ip': ip_address_prefix}
-            )
-            self.endpoint_2_rd[(mac_address, ip_address_prefix)] = endpoint_rd
+            self.localport_2_endpoints[linuxif].add(endpoint)
+            self.endpoint_2_lp[endpoint] = local_pref
+            self.endpoint_2_direction[endpoint] = direction
             self.mac_2_localport_data[mac_address] = pdata
 
-            if ip_address_prefix not in self.ip_address_2_mac:
-                self.ip_address_2_mac[ip_address_prefix] = list()
-
-            self.ip_address_2_mac[ip_address_prefix].append(mac_address)
+            self.ip_address_2_mac[ip_address_prefix].add(mac_address)
 
         except Exception as e:
             self.log.error("Error in vif_plugged: %s", e)
             if linuxif in self.localport_2_endpoints:
-                endpoint = {'mac': mac_address, 'ip': ip_address_prefix}
-                if endpoint in self.localport_2_endpoints[linuxif]:
-                    self.localport_2_endpoints[linuxif].remove(endpoint)
+                self.localport_2_endpoints[linuxif].discard(endpoint)
                 if not self.localport_2_endpoints[linuxif]:
                     del self.localport_2_endpoints[linuxif]
             if mac_address in self.mac_2_localport_data:
                 del self.mac_2_localport_data[mac_address]
-            if (ip_address_prefix in self.ip_address_2_mac and
-                    mac_address in self.ip_address_2_mac[ip_address_prefix]):
-                self.ip_address_2_mac[ip_address_prefix].remove(mac_address)
+            if ip_address_prefix in self.ip_address_2_mac:
+                self.ip_address_2_mac[ip_address_prefix].discard(mac_address)
 
             raise
 
         self.log.info("localport_2_endpoints: %s", self.localport_2_endpoints)
         self.log.info("endpoint_2_rd: %s", self.endpoint_2_rd)
-        self.log.info("mac_2_localport_data: %s",
-                      self.mac_2_localport_data)
+        self.log.info("mac_2_localport_data: %s", self.mac_2_localport_data)
         self.log.info("ip_address_2_mac: %s", self.ip_address_2_mac)
 
     @utils.synchronized
     @log_decorator.log_info
-    def vif_unplugged(self, mac_address, ip_address_prefix,
-                      advertise_subnet=False, lb_consistent_hash_order=0):
+    def vif_unplugged(self, mac_address, ip_address_prefix):
         # NOTE(tmorin): move this as a vif_unplugged_precheck, so that
         # in ipvpn.VRF this is done before readvertised route withdrawal
-
+        endpoint = (mac_address, ip_address_prefix)
         # Verify port and endpoint (MAC address, IP address) tuple consistency
         pdata = self.mac_2_localport_data.get(mac_address)
         if (not pdata or
                 (ip_address_prefix in self.ip_address_2_mac and
                  mac_address not in self.ip_address_2_mac[ip_address_prefix])
-                or
-                (mac_address, ip_address_prefix) not in self.endpoint_2_rd):
+                or endpoint not in self.endpoint_2_rd):
             self.log.error("vif_unplugged called for endpoint (%s, %s), but "
                            "no consistent informations or was not plugged yet",
-                           mac_address, ip_address_prefix)
+                           *endpoint)
             raise exc.APIException("Inconsistent endpoint (%s, %s) info "
                                    "or endpoint wasn't plugged yet, "
-                                   "cannot unplug" %
-                                   (mac_address, ip_address_prefix))
+                                   "cannot unplug" % endpoint)
 
         # Finding label and local port informations
         label = pdata.get('label')
-        endpoint_rd = self.endpoint_2_rd[(mac_address, ip_address_prefix)]
         localport = pdata.get('port_info')
         linuxif = localport['linuxif']
+        direction = self.endpoint_2_direction[endpoint]
         if not label or not localport:
             self.log.error("vif_unplugged called for endpoint (%s, %s), but "
                            "port data (%s, %s) is incomplete",
@@ -737,30 +819,20 @@ class VPNInstance(tracker_worker.TrackerWorker,
 
         if linuxif in self.localport_2_endpoints:
             # Parse address/mask
-            (ip_prefix, plen) = self._parse_ipaddress_prefix(ip_address_prefix)
+            (ip_prefix, _) = self._parse_ipaddress_prefix(ip_address_prefix)
 
-            last_endpoint = len(
-                self.localport_2_endpoints[linuxif]) <= 1
+            self.log.info("Withdrawing BGP route for VIF %s endpoint "
+                          "(%s, %s)", linuxif, mac_address, ip_address_prefix)
 
-            if not advertise_subnet and plen != 32:
-                self.log.debug("Using /32 instead of /%d", plen)
-                plen = 32
+            self._withdraw_route(self.endpoint_2_route.pop(endpoint))
 
-            rd = self.instance_rd if plen == 32 else endpoint_rd
+            last_endpoint = len(self.localport_2_endpoints[linuxif]) <= 1
 
-            self.log.info("Synthesizing and withdrawing BGP route for VIF %s "
-                          "endpoint (%s, %s/%d)", linuxif,
-                          mac_address, ip_prefix, plen)
-
-            route_entry = self.synthesize_vif_bgp_route(
-                mac_address, ip_prefix, plen,
-                label, lb_consistent_hash_order, rd)
-
-            self._withdraw_route(route_entry)
-
-            # Unplug endpoint from data plane
-            self.dataplane.vif_unplugged(
-                mac_address, ip_prefix, localport, label, last_endpoint)
+            if forward_to_port(direction):
+                # Unplug endpoint from data plane
+                self.dataplane.vif_unplugged(
+                    mac_address, ip_prefix, localport, label, direction,
+                    last_endpoint)
 
             # Forget data for this port if last endpoint
             if last_endpoint:
@@ -770,27 +842,28 @@ class VPNInstance(tracker_worker.TrackerWorker,
 
                 del self.localport_2_endpoints[linuxif]
             else:
-                self.localport_2_endpoints[linuxif].remove(
-                    {'mac': mac_address, 'ip': ip_address_prefix}
-                )
+                self.localport_2_endpoints[linuxif].remove(endpoint)
 
-            del self.endpoint_2_rd[(mac_address, ip_address_prefix)]
             # Free route distinguisher to the allocator
-            self.manager.rd_allocator.release(endpoint_rd)
+            if forward_to_port(direction):
+                self.manager.rd_allocator.release(
+                    self.endpoint_2_rd.pop(endpoint))
 
             if not last_endpoint:
-                if not any([endpoint['mac'] == mac_address
-                            for endpoint
+                if not any([ep[0] == mac_address
+                            for ep
                             in self.localport_2_endpoints[linuxif]]
                            ):
                     del self.mac_2_localport_data[mac_address]
             else:
                 del self.mac_2_localport_data[mac_address]
 
-            self.ip_address_2_mac[ip_address_prefix].remove(mac_address)
+            self.ip_address_2_mac[ip_address_prefix].discard(mac_address)
 
             if not self.ip_address_2_mac[ip_address_prefix]:
                 del self.ip_address_2_mac[ip_address_prefix]
+
+            del self.endpoint_2_desc[endpoint]
         else:
             self.log.error("vif_unplugged called for endpoint {%s, %s}, but"
                            " port data is incomplete", mac_address,
@@ -798,8 +871,7 @@ class VPNInstance(tracker_worker.TrackerWorker,
             raise Exception("bagpipe-bgp bug, check its logs")
 
         self.log.info("localport_2_endpoints: %s", self.localport_2_endpoints)
-        self.log.info("mac_2_localport_data: %s",
-                      self.mac_2_localport_data)
+        self.log.info("mac_2_localport_data: %s", self.mac_2_localport_data)
         self.log.info("ip_address_2_mac: %s", self.ip_address_2_mac)
 
     @utils.synchronized
@@ -815,7 +887,7 @@ class VPNInstance(tracker_worker.TrackerWorker,
     def stop_if_no_redirected_instance(self):
         self.log.debug("redirected_instances: %s", self.redirected_instances)
         if not self.redirected_instances:
-            self._stop()
+            self.stop()
             return True
 
         return False
@@ -913,19 +985,24 @@ class VPNInstance(tracker_worker.TrackerWorker,
         whether or not the route removed is the last one and depending on
         the desired behavior for the dataplane driver
         '''
-        return not last and not (self.dp_driver.makebefore4break_support or
-                                 self.dp_driver.ecmp_support)
+        if last:
+            # never skip the last route
+            return False
+
+        # if the driver supports ECMP, then it needs to see all the route
+        # withdraws
+        return not self.dp_driver.ecmp_support
 
     # Callbacks for BGP route updates (TrackerWorker) ########################
 
     @utils.synchronized
     @log_decorator.log
-    def _new_best_route(self, entry, new_route):
+    def new_best_route(self, entry, new_route):
         pass
 
     @utils.synchronized
     @log_decorator.log
-    def _best_route_removed(self, entry, old_route, last):
+    def best_route_removed(self, entry, old_route, last):
         pass
 
     # Looking Glass ####
@@ -934,6 +1011,7 @@ class VPNInstance(tracker_worker.TrackerWorker,
         return {
             "instance_type":         (lg.VALUE, self.instance_type),
             "external_instance_id":  (lg.VALUE, self.external_instance_id),
+            "description":           (lg.VALUE, self.description),
             "dataplane":             (lg.DELEGATE, self.dataplane),
             "route_targets":         (lg.SUBITEM, self.get_rts),
             "gateway_ip":            (lg.VALUE, self.gateway_ip),
@@ -941,6 +1019,7 @@ class VPNInstance(tracker_worker.TrackerWorker,
             "instance_dataplane_id": (lg.VALUE, self.instance_label),
             "ports":                 (lg.SUBTREE, self.get_lg_local_port_data),
             "readvertise":           (lg.SUBITEM, self.get_lg_readvertise),
+            "attract_traffic":       (lg.SUBITEM, self.get_lg_attract_traffic),
             "fallback":              (lg.VALUE, self.fallback)
         }
 
@@ -949,13 +1028,15 @@ class VPNInstance(tracker_worker.TrackerWorker,
         for (port, endpoints) in self.localport_2_endpoints.items():
             eps = []
             for endpoint in endpoints:
+                mac, ip = endpoint
                 eps.append({
-                    'label':
-                    self.mac_2_localport_data[endpoint['mac']]['label'],
-                    'mac_address': endpoint['mac'],
-                    'ip_address': endpoint['ip'],
-                    'rd': repr(self.endpoint_2_rd[(endpoint['mac'],
-                                                   endpoint['ip'])])
+                    'label': self.mac_2_localport_data[mac]['label'],
+                    'mac_address': mac,
+                    'ip_address': ip,
+                    'local_pref': self.endpoint_2_lp.get(endpoint),
+                    'rd': repr(self.endpoint_2_rd[endpoint]),
+                    'description': self.endpoint_2_desc.get(endpoint),
+                    'direction': self.endpoint_2_direction.get(endpoint)
                 })
 
             r[port] = {
@@ -972,12 +1053,35 @@ class VPNInstance(tracker_worker.TrackerWorker,
     def get_lg_readvertise(self):
         r = {}
         if self.readvertise:
-            r = {'from': [repr(rt) for rt in self.readvertise_from_rts],
-                 'to': [repr(rt) for rt in self.readvertise_to_rts]}
-
-            if self.attract_traffic:
-                r['attract_traffic'] = {
-                    'redirect_rts': [repr(rt) for rt in self.attract_rts],
-                    'classifier': self.attract_classifier
-                }
+            r = {
+                'from': [repr(rt) for rt in self.readvertise_from_rts],
+                'to': [repr(rt) for rt in self.readvertise_to_rts]
+            }
         return r
+
+    def get_lg_attract_traffic(self):
+        r = {}
+        if self.attract_traffic:
+            r = {
+                'redirect_rts': [repr(rt) for rt in self.attract_rts],
+                'classifier': self.attract_classifier,
+            }
+
+            if not hasattr(self, 'attract_static_dest_prefixes'):
+                return r
+
+            r.update({
+                'to': [repr(rt) for rt in self.readvertise_to_rts],
+                'static_destination_prefixes':
+                    self.attract_static_dest_prefixes
+            })
+
+        return r
+
+    def get_lg_summary(self):
+        # used from VPNManager looking glass
+        entry = {"id": self.external_instance_id,
+                 "name": self.name}
+        if self.description:
+            entry['description'] = self.description
+        return entry
