@@ -18,6 +18,8 @@
 import mock
 
 from networking_bagpipe.bagpipe_bgp import constants as consts
+
+from networking_bagpipe.bagpipe_bgp.common import utils
 from networking_bagpipe.bagpipe_bgp.vpn import manager
 from networking_bagpipe.tests.unit.bagpipe_bgp import base as t
 
@@ -68,7 +70,7 @@ class MockVPNInstance(object):
     def update_fallback(self, *args):
         pass
 
-    def vif_plugged(self, *args):
+    def vif_plugged(self, *args, **kwargs):
         pass
 
     def vif_unplugged(self, *args):
@@ -158,7 +160,7 @@ class TestVPNManager(t.TestCase):
 
     def test_plug_vif_to_vpn_with_forced_vni(self):
         with mock.patch.object(self.manager, "_get_vpn_instance") as mock_get_vpn_instance, \
-                mock.patch.object(manager, "convert_route_targets"):
+                mock.patch.object(utils, "convert_route_targets"):
             self.manager.plug_vif_to_vpn(vpn_instance_id=VPN_EXT_ID,
                                          vpn_type=consts.EVPN,
                                          import_rt=[t.RT1],
@@ -175,7 +177,7 @@ class TestVPNManager(t.TestCase):
 
     def test_plug_vif_to_vpn_without_forced_vni(self):
         with mock.patch.object(self.manager, "_get_vpn_instance") as mock_get_vpn_instance, \
-                mock.patch.object(manager, "convert_route_targets"):
+                mock.patch.object(utils, "convert_route_targets"):
             self.manager.plug_vif_to_vpn(vpn_instance_id=VPN_EXT_ID,
                                          vpn_type=consts.EVPN,
                                          import_rt=[t.RT1],
@@ -250,6 +252,33 @@ class TestVPNManager(t.TestCase):
             self.assertEqual(len(self.manager.vpn_instances.values()),
                              len(set(instance_ids)))
 
+    def test_instance_id_max(self):
+        with mock.patch.object(manager.VPNManager, 'type2class',
+                               {consts.IPVPN: MockVPNInstance,
+                                consts.EVPN: MockVPNInstance
+                                }):
+            self.manager.next_vpn_instance_id = 2**32 - 1
+
+            self.manager.plug_vif_to_vpn(
+                external_instance_id="dummy1",
+                vpn_type=consts.EVPN,
+                mac_address=MAC,
+                import_rts=[],
+                export_rts=[],
+                ip_address_prefix=IP,
+                local_port=LOCAL_PORT)
+
+            self.assertRaises(
+                manager.MaxInstanceIDReached,
+                self.manager.plug_vif_to_vpn,
+                external_instance_id="dummy2",
+                vpn_type=consts.EVPN,
+                import_rts=[],
+                export_rts=[],
+                mac_address=MAC,
+                ip_address_prefix=IP,
+                local_port=LOCAL_PORT)
+
     @mock.patch('networking_bagpipe.bagpipe_bgp.engine.bgp_manager.Manager')
     def test_manager_stop(self, mocked_bgp_manager):
         self.manager._get_vpn_instance("TEST_VPN_INSTANCE", consts.IPVPN,
