@@ -16,7 +16,7 @@
 # pip install {opts} {packages}
 
 ZUUL_CLONER=/usr/zuul-env/bin/zuul-cloner
-BRANCH_NAME=stable/pike
+BRANCH_NAME=master
 GIT_BASE=${GIT_BASE:-https://git.openstack.org/}
 
 install_project() {
@@ -30,7 +30,8 @@ install_project() {
 
     if [ $project_installed -eq 0 ]; then
         echo "ALREADY INSTALLED" > /tmp/tox_install.txt
-        echo "$project already installed; using existing package"
+        where=`python -c "import $module_name; print($module_name.__file__)"`
+        echo "$project already installed; using existing package: $where"
     elif [ -x "$ZUUL_CLONER" ]; then
         echo "ZUUL CLONER" > /tmp/tox_install.txt
         # Make this relative to current working directory so that
@@ -39,8 +40,7 @@ install_project() {
         mkdir -p .tmp
         PROJECT_DIR=$(/bin/mktemp -d -p $(pwd)/.tmp)
         pushd $PROJECT_DIR
-        $ZUUL_CLONER --cache-dir \
-            /opt/git \
+        $ZUUL_CLONER -v --cache-dir /opt/git \
             --branch $BRANCH_NAME \
             http://git.openstack.org \
             openstack/$project
@@ -63,6 +63,7 @@ shift
 
 install_project neutron
 install_project networking-bgpvpn
+install_project networking-sfc
 
 # install the rest of dependencies
 $install_cmd -U $*
